@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2014 Intel Corporation.
+ * Copyright © 2008-2015 Intel Corporation.
  * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
@@ -28,6 +28,9 @@
 #include <stdio.h>
 #include <sys/types.h>
 #include <stdarg.h>
+#ifdef HAVE_LZ4
+#include <lz4.h>
+#endif
 
 #include "openconnect-internal.h"
 
@@ -47,16 +50,22 @@ static const char data_hdr[8] = {
 	0		/* Unknown */
 };
 
+/* Strange initialisers here to work around GCC PR#10676 (which was
+ * fixed in GCC 4.6 but it takes a while for some systems to catch
+ * up. */
 static const struct pkt keepalive_pkt = {
-	.hdr = { 'S', 'T', 'F', 1, 0, 0, AC_PKT_KEEPALIVE, 0 },
+	.next = NULL,
+	{ .cstp.hdr = { 'S', 'T', 'F', 1, 0, 0, AC_PKT_KEEPALIVE, 0 } }
 };
 
 static const struct pkt dpd_pkt = {
-	.hdr = { 'S', 'T', 'F', 1, 0, 0, AC_PKT_DPD_OUT, 0 },
+	.next = NULL,
+	{ .cstp.hdr = { 'S', 'T', 'F', 1, 0, 0, AC_PKT_DPD_OUT, 0 } }
 };
 
 static const struct pkt dpd_resp_pkt = {
-	.hdr = { 'S', 'T', 'F', 1, 0, 0, AC_PKT_DPD_RESP, 0 },
+	.next = NULL,
+	{ .cstp.hdr = { 'S', 'T', 'F', 1, 0, 0, AC_PKT_DPD_RESP, 0 } }
 };
 
 /* Calculate MTU to request. Old servers simply use the X-CSTP-MTU: header,
@@ -120,36 +129,55 @@ static void calculate_mtu(struct openconnect_info *vpninfo, int *base_mtu, int *
 		*mtu = 1280;
 }
 
-void cstp_free_splits(struct openconnect_info *vpninfo)
-{
-	struct oc_split_include *inc;
-
-	for (inc = vpninfo->ip_info.split_includes; inc; ) {
-		struct oc_split_include *next = inc->next;
-		free(inc);
-		inc = next;
-	}
-	for (inc = vpninfo->ip_info.split_excludes; inc; ) {
-		struct oc_split_include *next = inc->next;
-		free(inc);
-		inc = next;
-	}
-	for (inc = vpninfo->ip_info.split_dns; inc; ) {
-		struct oc_split_include *next = inc->next;
-		free(inc);
-		inc = next;
-	}
-	vpninfo->ip_info.split_dns = vpninfo->ip_info.split_includes =
-		vpninfo->ip_info.split_excludes = NULL;
-}
-
-/* if DTLS 1.2 is supported */
+/* For OpenSSL the configure script detects DTLS 1.2 support.
+ * For GnuTLS just check for v3.2.0+ */
 #if defined(DTLS_GNUTLS) && GNUTLS_VERSION_NUMBER >= 0x030200
+#define HAVE_DTLS12 1
+#endif
+
+#ifdef HAVE_DTLS12
 # define DEFAULT_CIPHER_LIST "OC-DTLS1_2-AES256-GCM:OC-DTLS1_2-AES128-GCM:AES256-SHA:AES128-SHA:DES-CBC3-SHA:DES-CBC-SHA"
 #else
 # define DEFAULT_CIPHER_LIST "AES256-SHA:AES128-SHA:DES-CBC3-SHA:DES-CBC-SHA"
 #endif
 
+static void append_compr_types(struct oc_text_buf *buf, const char *proto, int avail)
+{
+	if (avail) {
+		char sep = ' ';
+		buf_append(buf, "X-%s-Accept-Encoding:", proto);
+		if (avail & COMPR_LZ4) {
+			buf_append(buf, "%coc-lz4", sep);
+			sep = ',';
+		}
+		if (avail & COMPR_LZS) {
+			buf_append(buf, "%clzs", sep);
+			sep = ',';
+		}
+		if (avail & COMPR_DEFLATE) {
+			buf_append(buf, "%cdeflate", sep);
+			sep = ',';
+		}
+		buf_append(buf, "\r\n");
+	}
+}
+
+static void append_mobile_headers(struct openconnect_info *vpninfo, struct oc_text_buf *buf)
+{
+	if (vpninfo->mobile_platform_version) {
+		buf_append(buf, "X-AnyConnect-Identifier-ClientVersion: %s\r\n",
+			   openconnect_version_str);
+		buf_append(buf, "X-AnyConnect-Identifier-Platform: %s\r\n",
+			   vpninfo->platname);
+		buf_append(buf, "X-AnyConnect-Identifier-PlatformVersion: %s\r\n",
+			   vpninfo->mobile_platform_version);
+		buf_append(buf, "X-AnyConnect-Identifier-DeviceType: %s\r\n",
+			   vpninfo->mobile_device_type);
+		buf_append(buf, "X-AnyConnect-Identifier-Device-UniqueID: %s\r\n",
+			   vpninfo->mobile_device_uniqueid);
+	}
+}
+
 static int start_cstp_connection(struct openconnect_info *vpninfo)
 {
 	struct oc_text_buf *reqbuf;
@@ -176,7 +204,7 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 
 	for (i = 0; i < 3; i++)
 		vpninfo->ip_info.dns[i] = vpninfo->ip_info.nbns[i] = NULL;
-	cstp_free_splits(vpninfo);
+	free_split_routes(vpninfo);
 
  retry:
 	calculate_mtu(vpninfo, &base_mtu, &mtu);
@@ -188,15 +216,10 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	buf_append(reqbuf, "Cookie: webvpn=%s\r\n", vpninfo->cookie);
 	buf_append(reqbuf, "X-CSTP-Version: 1\r\n");
 	buf_append(reqbuf, "X-CSTP-Hostname: %s\r\n", vpninfo->localname);
-	if (vpninfo->req_compr) {
-		char sep = ' ';
-		buf_append(reqbuf, "X-CSTP-Accept-Encoding:");
-		if (vpninfo->req_compr & COMPR_DEFLATE) {
-			buf_append(reqbuf, "%cdeflate", sep);
-			sep = ',';
-		}
-		buf_append(reqbuf, "\r\n");
-	}
+
+	append_mobile_headers(vpninfo, reqbuf);
+	append_compr_types(reqbuf, "CSTP", vpninfo->req_compr);
+
 	if (base_mtu)
 		buf_append(reqbuf, "X-CSTP-Base-MTU: %d\r\n", base_mtu);
 	buf_append(reqbuf, "X-CSTP-MTU: %d\r\n", mtu);
@@ -204,25 +227,34 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 			       vpninfo->disable_ipv6 ? "IPv4" : "IPv6,IPv4");
 	if (!vpninfo->disable_ipv6)
 		buf_append(reqbuf, "X-CSTP-Full-IPv6-Capability: true\r\n");
-	buf_append(reqbuf, "X-DTLS-Master-Secret: ");
-	for (i = 0; i < sizeof(vpninfo->dtls_secret); i++) {
-		buf_append(reqbuf, "%02X", vpninfo->dtls_secret[i]);
-		dtls_secret_set |= vpninfo->dtls_secret[i];
-	}
-	if (!dtls_secret_set) {
-		vpn_progress(vpninfo, PRG_ERR,
-			     _("CRITICAL ERROR: DTLS master secret is uninitialised. Please report this.\n"));
-		return -EINVAL;
-	}
+	if (vpninfo->dtls_state != DTLS_DISABLED) {
+		buf_append(reqbuf, "X-DTLS-Master-Secret: ");
+		for (i = 0; i < sizeof(vpninfo->dtls_secret); i++) {
+			buf_append(reqbuf, "%02X", vpninfo->dtls_secret[i]);
+			dtls_secret_set |= vpninfo->dtls_secret[i];
+		}
+		if (!dtls_secret_set) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("CRITICAL ERROR: DTLS master secret is uninitialised. Please report this.\n"));
+			buf_free(reqbuf);
+			return -EINVAL;
+		}
+		buf_append(reqbuf, "\r\nX-DTLS-CipherSuite: %s\r\n",
+			   vpninfo->dtls_ciphers ? : DEFAULT_CIPHER_LIST);
 
-	buf_append(reqbuf, "\r\nX-DTLS-CipherSuite: %s\r\n\r\n",
-			       vpninfo->dtls_ciphers ? : DEFAULT_CIPHER_LIST);
+		append_compr_types(reqbuf, "DTLS", vpninfo->req_compr & ~COMPR_DEFLATE);
+	}
+	buf_append(reqbuf, "\r\n");
 
 	if (buf_error(reqbuf)) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Error creating HTTPS CONNECT request\n"));
 		return buf_free(reqbuf);
 	}
+
+	if (vpninfo->dump_http_traffic)
+		dump_buf(vpninfo, '>', reqbuf->data);
+
 	vpninfo->ssl_write(vpninfo, reqbuf->data, reqbuf->pos);
 	buf_free(reqbuf);
 
@@ -349,6 +381,17 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 
 				if (dtls_sessid_changed && vpninfo->dtls_state > DTLS_SLEEPING)
 					vpninfo->dtls_need_reconnect = 1;
+			} else if (!strcmp(buf + 7, "Content-Encoding")) {
+				if (!strcmp(colon, "lzs"))
+					vpninfo->dtls_compr = COMPR_LZS;
+				else if (!strcmp(colon, "oc-lz4"))
+					vpninfo->dtls_compr = COMPR_LZ4;
+				else {
+					vpn_progress(vpninfo, PRG_ERR,
+						     _("Unknown DTLS-Content-Encoding %s\n"),
+						     colon);
+					return -EINVAL;
+				}
 			}
 			continue;
 		}
@@ -375,6 +418,10 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 		} else if (!strcmp(buf + 7, "Content-Encoding")) {
 			if (!strcmp(colon, "deflate"))
 				vpninfo->cstp_compr = COMPR_DEFLATE;
+			else if (!strcmp(colon, "lzs"))
+				vpninfo->cstp_compr = COMPR_LZS;
+			else if (!strcmp(colon, "oc-lz4"))
+				vpninfo->cstp_compr = COMPR_LZ4;
 			else {
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("Unknown CSTP-Content-Encoding %s\n"),
@@ -535,10 +582,11 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 }
 
 
-int openconnect_make_cstp_connection(struct openconnect_info *vpninfo)
+int cstp_connect(struct openconnect_info *vpninfo)
 {
 	int ret;
 	int deflate_bufsize = 0;
+	int compr_type;
 
 	/* This needs to be done before openconnect_setup_dtls() because it's
 	   sent with the CSTP CONNECT handshake. Even if we don't end up doing
@@ -547,7 +595,7 @@ int openconnect_make_cstp_connection(struct openconnect_info *vpninfo)
 		if (openconnect_random(vpninfo->dtls_secret, sizeof(vpninfo->dtls_secret)))
 			return -EINVAL;
 		/* The application will later call openconnect_setup_dtls() */
-		vpninfo->dtls_state = DTLS_DISABLED;
+		vpninfo->dtls_state = DTLS_SECRET;
 	}
 
 	ret = openconnect_open_https(vpninfo);
@@ -558,9 +606,18 @@ int openconnect_make_cstp_connection(struct openconnect_info *vpninfo)
 	if (ret)
 		goto out;
 
+	/* Allow for the theoretical possibility of having *different*
+	 * compression type for CSTP and DTLS. Although all we've seen
+	 * in practice is that one is enabled and the other isn't. */
+	compr_type = vpninfo->cstp_compr | vpninfo->dtls_compr;
+
+	/* This will definitely be smaller than zlib's */
+	if (compr_type & (COMPR_LZS|COMPR_LZ4))
+		deflate_bufsize = vpninfo->ip_info.mtu;
+
 	/* If deflate compression is enabled (which is CSTP-only), it needs its
 	 * context to be allocated. */
-	if (vpninfo->cstp_compr == COMPR_DEFLATE) {
+	if (compr_type & COMPR_DEFLATE) {
 		vpninfo->deflate_adler32 = 1;
 		vpninfo->inflate_adler32 = 1;
 
@@ -591,8 +648,8 @@ int openconnect_make_cstp_connection(struct openconnect_info *vpninfo)
 
 		vpninfo->deflate_pkt_size = deflate_bufsize;
 		memset(vpninfo->deflate_pkt, 0, sizeof(struct pkt));
-		memcpy(vpninfo->deflate_pkt->hdr, data_hdr, 8);
-		vpninfo->deflate_pkt->hdr[6] = AC_PKT_COMPRESSED;
+		memcpy(vpninfo->deflate_pkt->cstp.hdr, data_hdr, 8);
+		vpninfo->deflate_pkt->cstp.hdr[6] = AC_PKT_COMPRESSED;
 	}
 
  out:
@@ -604,12 +661,6 @@ int openconnect_make_cstp_connection(struct openconnect_info *vpninfo)
 
 static int cstp_reconnect(struct openconnect_info *vpninfo)
 {
-	int ret;
-	int timeout;
-	int interval;
-
-	openconnect_close_https(vpninfo, 0);
-
 	if (vpninfo->cstp_compr == COMPR_DEFLATE) {
 		/* Requeue the original packet that was deflated */
 		if (vpninfo->current_ssl_pkt == vpninfo->deflate_pkt) {
@@ -620,163 +671,150 @@ static int cstp_reconnect(struct openconnect_info *vpninfo)
 		inflateEnd(&vpninfo->inflate_strm);
 		deflateEnd(&vpninfo->deflate_strm);
 	}
-	timeout = vpninfo->reconnect_timeout;
-	interval = vpninfo->reconnect_interval;
 
-	free(vpninfo->dtls_pkt);
-	vpninfo->dtls_pkt = NULL;
-	free(vpninfo->tun_pkt);
-	vpninfo->tun_pkt = NULL;
-
-	while ((ret = openconnect_make_cstp_connection(vpninfo))) {
-		if (timeout <= 0)
-			return ret;
-		if (ret == -EPERM) {
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Cookie is no longer valid, ending session\n"));
-			return ret;
-		}
-		vpn_progress(vpninfo, PRG_INFO,
-			     _("sleep %ds, remaining timeout %ds\n"),
-			     interval, timeout);
-		poll_cmd_fd(vpninfo, interval);
-		if (vpninfo->got_cancel_cmd)
-			return -EINTR;
-		if (vpninfo->got_pause_cmd)
-			return 0;
-		timeout -= interval;
-		interval += vpninfo->reconnect_interval;
-		if (interval > RECONNECT_INTERVAL_MAX)
-			interval = RECONNECT_INTERVAL_MAX;
-	}
-	script_config_tun(vpninfo, "reconnect");
-	return 0;
+	return ssl_reconnect(vpninfo);
 }
 
-static int inflate_and_queue_packet(struct openconnect_info *vpninfo,
-				    unsigned char *buf, int len)
+int decompress_and_queue_packet(struct openconnect_info *vpninfo, int compr_type,
+				unsigned char *buf, int len)
 {
 	struct pkt *new = malloc(sizeof(struct pkt) + vpninfo->ip_info.mtu);
-	uint32_t pkt_sum;
+	const char *comprname = "";
 
 	if (!new)
 		return -ENOMEM;
 
 	new->next = NULL;
 
-	vpninfo->inflate_strm.next_in = buf;
-	vpninfo->inflate_strm.avail_in = len - 4;
+	if (compr_type == COMPR_DEFLATE) {
+		uint32_t pkt_sum;
 
-	vpninfo->inflate_strm.next_out = new->data;
-	vpninfo->inflate_strm.avail_out = vpninfo->ip_info.mtu;
-	vpninfo->inflate_strm.total_out = 0;
+		comprname = "deflate";
 
-	if (inflate(&vpninfo->inflate_strm, Z_SYNC_FLUSH)) {
-		vpn_progress(vpninfo, PRG_ERR, _("inflate failed\n"));
+		vpninfo->inflate_strm.next_in = buf;
+		vpninfo->inflate_strm.avail_in = len - 4;
+
+		vpninfo->inflate_strm.next_out = new->data;
+		vpninfo->inflate_strm.avail_out = vpninfo->ip_info.mtu;
+		vpninfo->inflate_strm.total_out = 0;
+
+		if (inflate(&vpninfo->inflate_strm, Z_SYNC_FLUSH)) {
+			vpn_progress(vpninfo, PRG_ERR, _("inflate failed\n"));
+			free(new);
+			return -EINVAL;
+		}
+
+		new->len = vpninfo->inflate_strm.total_out;
+
+		vpninfo->inflate_adler32 = adler32(vpninfo->inflate_adler32,
+						   new->data, new->len);
+
+		pkt_sum = load_be32(buf + len - 4);
+
+		if (vpninfo->inflate_adler32 != pkt_sum)
+			vpninfo->quit_reason = "Compression (inflate) adler32 failure";
+
+	} else if (compr_type == COMPR_LZS) {
+		comprname = "LZS";
+
+		new->len = lzs_decompress(new->data, vpninfo->ip_info.mtu, buf, len);
+		if (new->len < 0) {
+			len = new->len;
+			if (len == 0)
+				len = -EINVAL;
+			vpn_progress(vpninfo, PRG_ERR, _("LZS decompression failed: %s\n"),
+				     strerror(-len));
+			free(new);
+			return len;
+		}
+#ifdef HAVE_LZ4
+	} else if (compr_type == COMPR_LZ4) {
+		comprname = "LZ4";
+		new->len = LZ4_decompress_safe((void *)buf, (void *)new->data, len, vpninfo->ip_info.mtu);
+		if (new->len <= 0) {
+			len = new->len;
+			if (len == 0)
+				len = -EINVAL;
+			vpn_progress(vpninfo, PRG_ERR, _("LZ4 decompression failed\n"));
+			free(new);
+			return len;
+		}
+#endif
+	} else {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Unknown compression type %d\n"), compr_type);
 		free(new);
 		return -EINVAL;
 	}
-
-	new->len = vpninfo->inflate_strm.total_out;
-
-	vpninfo->inflate_adler32 = adler32(vpninfo->inflate_adler32,
-					   new->data, new->len);
-
-	pkt_sum = buf[len - 1] | (buf[len - 2] << 8) |
-		(buf[len - 3] << 16) | (buf[len - 4] << 24);
-
-	if (vpninfo->inflate_adler32 != pkt_sum)
-		vpninfo->quit_reason = "Compression (inflate) adler32 failure";
-
 	vpn_progress(vpninfo, PRG_TRACE,
-		     _("Received compressed data packet of %ld bytes\n"),
-		     (long)vpninfo->inflate_strm.total_out);
+		     _("Received %s compressed data packet of %d bytes (was %d)\n"),
+		     comprname, new->len, len);
 
 	queue_packet(&vpninfo->incoming_queue, new);
 	return 0;
 }
 
-#if defined(OPENCONNECT_OPENSSL)
-static int cstp_read(struct openconnect_info *vpninfo, void *buf, int maxlen)
-{
-	int len, ret;
-
-	len = SSL_read(vpninfo->https_ssl, buf, maxlen);
-	if (len > 0)
-		return len;
-
-	ret = SSL_get_error(vpninfo->https_ssl, len);
-	if (ret == SSL_ERROR_SYSCALL || ret == SSL_ERROR_ZERO_RETURN) {
-		vpn_progress(vpninfo, PRG_ERR,
-			     _("SSL read error %d (server probably closed connection); reconnecting.\n"),
-			     ret);
-		return -EIO;
-	}
-	return 0;
-}
-
-static int cstp_write(struct openconnect_info *vpninfo, void *buf, int buflen)
+int compress_packet(struct openconnect_info *vpninfo, int compr_type, struct pkt *this)
 {
 	int ret;
 
-	ret = SSL_write(vpninfo->https_ssl, buf, buflen);
-	if (ret > 0)
-		return ret;
+	if (compr_type == COMPR_DEFLATE) {
+		vpninfo->deflate_strm.next_in = this->data;
+		vpninfo->deflate_strm.avail_in = this->len;
+		vpninfo->deflate_strm.next_out = (void *)vpninfo->deflate_pkt->data;
+		vpninfo->deflate_strm.avail_out = vpninfo->deflate_pkt_size - 4;
+		vpninfo->deflate_strm.total_out = 0;
 
-	ret = SSL_get_error(vpninfo->https_ssl, ret);
-	switch (ret) {
-	case SSL_ERROR_WANT_WRITE:
-		/* Waiting for the socket to become writable -- it's
-		   probably stalled, and/or the buffers are full */
-		monitor_write_fd(vpninfo, ssl);
-	case SSL_ERROR_WANT_READ:
-		return 0;
-
-	default:
-		vpn_progress(vpninfo, PRG_ERR, _("SSL_write failed: %d\n"), ret);
-		openconnect_report_ssl_errors(vpninfo);
-		return -1;
-	}
-}
-#elif defined(OPENCONNECT_GNUTLS)
-static int cstp_read(struct openconnect_info *vpninfo, void *buf, int maxlen)
-{
-	int ret;
-
-	ret = gnutls_record_recv(vpninfo->https_sess, buf, maxlen);
-	if (ret > 0)
-		return ret;
-
-	if (ret != GNUTLS_E_AGAIN) {
-		vpn_progress(vpninfo, PRG_ERR,
-			     _("SSL read error: %s; reconnecting.\n"),
-			     gnutls_strerror(ret));
-		return -EIO;
-	}
-	return 0;
-}
-
-static int cstp_write(struct openconnect_info *vpninfo, void *buf, int buflen)
-{
-	int ret;
-
-	ret = gnutls_record_send(vpninfo->https_sess, buf, buflen);
-	if (ret > 0)
-		return ret;
-
-	if (ret == GNUTLS_E_AGAIN) {
-		if (gnutls_record_get_direction(vpninfo->https_sess)) {
-			/* Waiting for the socket to become writable -- it's
-			   probably stalled, and/or the buffers are full */
-			monitor_write_fd(vpninfo, ssl);
+		ret = deflate(&vpninfo->deflate_strm, Z_SYNC_FLUSH);
+		if (ret) {
+			vpn_progress(vpninfo, PRG_ERR, _("deflate failed %d\n"), ret);
+			/* Things are going to go horribly wrong if we try to do any
+			   more compression. Give up entirely. */
+			vpninfo->cstp_compr = 0;
+			return -EIO;
 		}
+
+		/* Add ongoing adler32 to tail of compressed packet */
+		vpninfo->deflate_adler32 = adler32(vpninfo->deflate_adler32,
+						   this->data, this->len);
+
+		store_be32(&vpninfo->deflate_pkt->data[vpninfo->deflate_strm.total_out],
+			   vpninfo->deflate_adler32);
+
+		vpninfo->deflate_pkt->len = vpninfo->deflate_strm.total_out + 4;
+		return 0;
+	} else if (compr_type == COMPR_LZS) {
+		if (this->len < 40)
+			return -EFBIG;
+
+		ret = lzs_compress(vpninfo->deflate_pkt->data, this->len,
+				   this->data, this->len);
+		if (ret < 0)
+			return ret;
+
+		vpninfo->deflate_pkt->len = ret;
+		return 0;
+#ifdef HAVE_LZ4
+	} else if (compr_type == COMPR_LZ4) {
+		if (this->len < 40)
+			return -EFBIG;
+
+		ret = LZ4_compress_limitedOutput((void*)this->data, (void*)vpninfo->deflate_pkt->data, this->len,
+				   this->len);
+		if (ret <= 0) {
+			if (ret == 0)
+				ret = -EFBIG;
+			return ret;
+		}
+
+		vpninfo->deflate_pkt->len = ret;
 		return 0;
-	}
-	vpn_progress(vpninfo, PRG_ERR, _("SSL send failed: %s\n"),
-		     gnutls_strerror(ret));
-	return -1;
-}
 #endif
+	} else
+		return -EINVAL;
+
+	return 0;
+}
 
 int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 {
@@ -804,7 +842,7 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 			}
 		}
 
-		len = cstp_read(vpninfo, vpninfo->cstp_pkt->hdr, len + 8);
+		len = ssl_nonblock_read(vpninfo, vpninfo->cstp_pkt->cstp.hdr, len + 8);
 		if (!len)
 			break;
 		if (len < 0)
@@ -815,26 +853,26 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 			return 1;
 		}
 
-		if (vpninfo->cstp_pkt->hdr[0] != 'S' || vpninfo->cstp_pkt->hdr[1] != 'T' ||
-		    vpninfo->cstp_pkt->hdr[2] != 'F' || vpninfo->cstp_pkt->hdr[3] != 1 ||
-		    vpninfo->cstp_pkt->hdr[7])
+		if (vpninfo->cstp_pkt->cstp.hdr[0] != 'S' || vpninfo->cstp_pkt->cstp.hdr[1] != 'T' ||
+		    vpninfo->cstp_pkt->cstp.hdr[2] != 'F' || vpninfo->cstp_pkt->cstp.hdr[3] != 1 ||
+		    vpninfo->cstp_pkt->cstp.hdr[7])
 			goto unknown_pkt;
 
-		payload_len = (vpninfo->cstp_pkt->hdr[4] << 8) + vpninfo->cstp_pkt->hdr[5];
+		payload_len = load_be16(vpninfo->cstp_pkt->cstp.hdr + 4);
 		if (len != 8 + payload_len) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Unexpected packet length. SSL_read returned %d but packet is\n"),
 				     len);
 			vpn_progress(vpninfo, PRG_ERR,
 				     "%02x %02x %02x %02x %02x %02x %02x %02x\n",
-				     vpninfo->cstp_pkt->hdr[0], vpninfo->cstp_pkt->hdr[1],
-				     vpninfo->cstp_pkt->hdr[2], vpninfo->cstp_pkt->hdr[3],
-				     vpninfo->cstp_pkt->hdr[4], vpninfo->cstp_pkt->hdr[5],
-				     vpninfo->cstp_pkt->hdr[6], vpninfo->cstp_pkt->hdr[7]);
+				     vpninfo->cstp_pkt->cstp.hdr[0], vpninfo->cstp_pkt->cstp.hdr[1],
+				     vpninfo->cstp_pkt->cstp.hdr[2], vpninfo->cstp_pkt->cstp.hdr[3],
+				     vpninfo->cstp_pkt->cstp.hdr[4], vpninfo->cstp_pkt->cstp.hdr[5],
+				     vpninfo->cstp_pkt->cstp.hdr[6], vpninfo->cstp_pkt->cstp.hdr[7]);
 			continue;
 		}
 		vpninfo->ssl_times.last_rx = time(NULL);
-		switch (vpninfo->cstp_pkt->hdr[6]) {
+		switch (vpninfo->cstp_pkt->cstp.hdr[6]) {
 		case AC_PKT_DPD_OUT:
 			vpn_progress(vpninfo, PRG_DEBUG,
 				     _("Got CSTP DPD request\n"));
@@ -863,14 +901,18 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 
 		case AC_PKT_DISCONN: {
 			int i;
-			for (i = 1; i < payload_len; i++) {
-				if (!isprint(vpninfo->cstp_pkt->data[i]))
-					vpninfo->cstp_pkt->data[i] = '.';
+			if (payload_len >= 2) {
+				for (i = 1; i < payload_len; i++) {
+					if (!isprint(vpninfo->cstp_pkt->data[i]))
+						vpninfo->cstp_pkt->data[i] = '.';
+				}
+				vpninfo->cstp_pkt->data[payload_len] = 0;
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Received server disconnect: %02x '%s'\n"),
+					     vpninfo->cstp_pkt->data[0], vpninfo->cstp_pkt->data + 1);
+			} else {
+				vpn_progress(vpninfo, PRG_ERR, _("Received server disconnect\n"));
 			}
-			vpninfo->cstp_pkt->data[payload_len] = 0;
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Received server disconnect: %02x '%s'\n"),
-				     vpninfo->cstp_pkt->data[0], vpninfo->cstp_pkt->data + 1);
 			vpninfo->quit_reason = "Server request";
 			return -EPIPE;
 		}
@@ -880,7 +922,8 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 					     _("Compressed packet received in !deflate mode\n"));
 				goto unknown_pkt;
 			}
-			inflate_and_queue_packet(vpninfo, vpninfo->cstp_pkt->data, payload_len);
+			decompress_and_queue_packet(vpninfo, vpninfo->cstp_compr,
+						    vpninfo->cstp_pkt->data, payload_len);
 			work_done = 1;
 			continue;
 
@@ -893,10 +936,10 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	unknown_pkt:
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Unknown packet %02x %02x %02x %02x %02x %02x %02x %02x\n"),
-			     vpninfo->cstp_pkt->hdr[0], vpninfo->cstp_pkt->hdr[1],
-			     vpninfo->cstp_pkt->hdr[2], vpninfo->cstp_pkt->hdr[3],
-			     vpninfo->cstp_pkt->hdr[4], vpninfo->cstp_pkt->hdr[5],
-			     vpninfo->cstp_pkt->hdr[6], vpninfo->cstp_pkt->hdr[7]);
+			     vpninfo->cstp_pkt->cstp.hdr[0], vpninfo->cstp_pkt->cstp.hdr[1],
+			     vpninfo->cstp_pkt->cstp.hdr[2], vpninfo->cstp_pkt->cstp.hdr[3],
+			     vpninfo->cstp_pkt->cstp.hdr[4], vpninfo->cstp_pkt->cstp.hdr[5],
+			     vpninfo->cstp_pkt->cstp.hdr[6], vpninfo->cstp_pkt->cstp.hdr[7]);
 		vpninfo->quit_reason = "Unknown packet received";
 		return 1;
 	}
@@ -910,16 +953,16 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 		vpninfo->ssl_times.last_tx = time(NULL);
 		unmonitor_write_fd(vpninfo, ssl);
 
-		ret = cstp_write(vpninfo,
-				 vpninfo->current_ssl_pkt->hdr,
-				 vpninfo->current_ssl_pkt->len + 8);
+		ret = ssl_nonblock_write(vpninfo,
+					 vpninfo->current_ssl_pkt->cstp.hdr,
+					 vpninfo->current_ssl_pkt->len + 8);
 		if (ret < 0)
 			goto do_reconnect;
 		else if (!ret) {
-			/* -EAGAIN: cstp_write() will have added the SSL fd to
-			   ->select_wfds if appropriate, so we can just return
-			   and wait. Unless it's been stalled for so long that
-			   DPD kicks in and we kill the connection. */
+			/* -EAGAIN: ssl_nonblock_write() will have added the SSL
+			   fd to ->select_wfds if appropriate, so we can just
+			   return and wait. Unless it's been stalled for so long
+			   that DPD kicks in and we kill the connection. */
 			switch (ka_stalled_action(&vpninfo->ssl_times, timeout)) {
 			case KA_DPD_DEAD:
 				goto peer_dead;
@@ -941,9 +984,10 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 			return 1;
 		}
 		/* Don't free the 'special' packets */
-		if (vpninfo->current_ssl_pkt == vpninfo->deflate_pkt)
+		if (vpninfo->current_ssl_pkt == vpninfo->deflate_pkt) {
 			free(vpninfo->pending_deflated_pkt);
-		else if (vpninfo->current_ssl_pkt != &dpd_pkt &&
+			vpninfo->pending_deflated_pkt = NULL;
+		} else if (vpninfo->current_ssl_pkt != &dpd_pkt &&
 			 vpninfo->current_ssl_pkt != &dpd_resp_pkt &&
 			 vpninfo->current_ssl_pkt != &keepalive_pkt)
 			free(vpninfo->current_ssl_pkt);
@@ -1008,7 +1052,8 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	case KA_KEEPALIVE:
 		/* No need to send an explicit keepalive
 		   if we have real data to send */
-		if (vpninfo->dtls_state != DTLS_CONNECTED && vpninfo->outgoing_queue)
+		if (vpninfo->dtls_state != DTLS_CONNECTED &&
+		    vpninfo->outgoing_queue.head)
 			break;
 
 		vpn_progress(vpninfo, PRG_DEBUG, _("Send CSTP Keepalive\n"));
@@ -1021,53 +1066,30 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	}
 
 	/* Service outgoing packet queue, if no DTLS */
-	while (vpninfo->dtls_state != DTLS_CONNECTED && vpninfo->outgoing_queue) {
-		struct pkt *this = vpninfo->outgoing_queue;
-		vpninfo->outgoing_queue = this->next;
-		vpninfo->outgoing_qlen--;
+	while (vpninfo->dtls_state != DTLS_CONNECTED &&
+	       (vpninfo->current_ssl_pkt = dequeue_packet(&vpninfo->outgoing_queue))) {
+		struct pkt *this = vpninfo->current_ssl_pkt;
 
-		if (vpninfo->cstp_compr == COMPR_DEFLATE) {
-			unsigned char *adler;
-			int ret;
-
-			vpninfo->deflate_strm.next_in = this->data;
-			vpninfo->deflate_strm.avail_in = this->len;
-			vpninfo->deflate_strm.next_out = (void *)vpninfo->deflate_pkt->data;
-			vpninfo->deflate_strm.avail_out = vpninfo->deflate_pkt_size - 4;
-			vpninfo->deflate_strm.total_out = 0;
-
-			ret = deflate(&vpninfo->deflate_strm, Z_SYNC_FLUSH);
-			if (ret) {
-				vpn_progress(vpninfo, PRG_ERR, _("deflate failed %d\n"), ret);
+		if (vpninfo->cstp_compr) {
+			ret = compress_packet(vpninfo, vpninfo->cstp_compr, this);
+			if (ret < 0)
 				goto uncompr;
-			}
 
-			vpninfo->deflate_pkt->hdr[4] = (vpninfo->deflate_strm.total_out + 4) >> 8;
-			vpninfo->deflate_pkt->hdr[5] = (vpninfo->deflate_strm.total_out + 4) & 0xff;
+			store_be16(vpninfo->deflate_pkt->cstp.hdr + 4, vpninfo->deflate_pkt->len);
 
-			/* Add ongoing adler32 to tail of compressed packet */
-			vpninfo->deflate_adler32 = adler32(vpninfo->deflate_adler32,
-							   this->data, this->len);
-
-			adler = &vpninfo->deflate_pkt->data[vpninfo->deflate_strm.total_out];
-			*(adler++) =  vpninfo->deflate_adler32 >> 24;
-			*(adler++) = (vpninfo->deflate_adler32 >> 16) & 0xff;
-			*(adler++) = (vpninfo->deflate_adler32 >> 8) & 0xff;
-			*(adler)   =  vpninfo->deflate_adler32 & 0xff;
-
-			vpninfo->deflate_pkt->len = vpninfo->deflate_strm.total_out + 4;
+			/* DTLS compression may have screwed with this */
+			vpninfo->deflate_pkt->cstp.hdr[7] = 0;
 
 			vpn_progress(vpninfo, PRG_TRACE,
-				     _("Sending compressed data packet of %d bytes\n"),
-				     this->len);
+				     _("Sending compressed data packet of %d bytes (was %d)\n"),
+				     vpninfo->deflate_pkt->len, this->len);
 
 			vpninfo->pending_deflated_pkt = this;
 			vpninfo->current_ssl_pkt = vpninfo->deflate_pkt;
 		} else {
 		uncompr:
-			memcpy(this->hdr, data_hdr, 8);
-			this->hdr[4] = this->len >> 8;
-			this->hdr[5] = this->len & 0xff;
+			memcpy(this->cstp.hdr, data_hdr, 8);
+			store_be16(this->cstp.hdr + 4, this->len);
 
 			vpn_progress(vpninfo, PRG_TRACE,
 				     _("Sending uncompressed data packet of %d bytes\n"),
@@ -1104,16 +1126,33 @@ int cstp_bye(struct openconnect_info *vpninfo, const char *reason)
 	memcpy(bye_pkt, data_hdr, 8);
 	memcpy(bye_pkt + 9, reason, reason_len);
 
-	bye_pkt[4] = (reason_len + 1) >> 8;
-	bye_pkt[5] = (reason_len + 1) & 0xff;
+	store_be16(bye_pkt + 4, reason_len + 1);
 	bye_pkt[6] = AC_PKT_DISCONN;
 	bye_pkt[8] = 0xb0;
 
 	vpn_progress(vpninfo, PRG_INFO,
 		     _("Send BYE packet: %s\n"), reason);
 
-	cstp_write(vpninfo, bye_pkt, reason_len + 9);
+	ssl_nonblock_write(vpninfo, bye_pkt, reason_len + 9);
 	free(bye_pkt);
 
 	return 0;
 }
+
+void cstp_common_headers(struct openconnect_info *vpninfo, struct oc_text_buf *buf)
+{
+	http_common_headers(vpninfo, buf);
+
+	buf_append(buf, "Accept: */*\r\n");
+	buf_append(buf, "Accept-Encoding: identity\r\n");
+	buf_append(buf, "X-Transcend-Version: 1\r\n");
+	if (vpninfo->xmlpost) {
+		buf_append(buf, "X-Aggregate-Auth: 1\r\n");
+		buf_append(buf, "X-AnyConnect-Platform: %s\r\n",
+			   vpninfo->platname);
+	}
+	if (vpninfo->try_http_auth)
+		buf_append(buf, "X-Support-HTTP-Auth: true\r\n");
+
+	append_mobile_headers(vpninfo, buf);
+}
