@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2014 Intel Corporation.
+ * Copyright © 2008-2015 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -28,19 +28,6 @@
 
 #include "openconnect-internal.h"
 
-static unsigned char nybble(unsigned char n)
-{
-	if      (n >= '0' && n <= '9') return n - '0';
-	else if (n >= 'A' && n <= 'F') return n - ('A' - 10);
-	else if (n >= 'a' && n <= 'f') return n - ('a' - 10);
-	return 0;
-}
-
-unsigned char unhex(const char *data)
-{
-	return (nybble(data[0]) << 4) | nybble(data[1]);
-}
-
 #ifdef HAVE_DTLS
 
 #if 0
@@ -103,22 +90,166 @@ int RAND_bytes(char *buf, int len)
 #define DTLS_RECV SSL_read
 #define DTLS_FREE SSL_free
 
+/* In the very early days there were cases where this wasn't found in
+ * the header files but it did still work somehow. I forget the details
+ * now but I was definitely avoiding using the macro. Let's just define
+ * it for ourselves instead.*/
+#ifndef DTLS1_BAD_VER
+#define DTLS1_BAD_VER 0x100
+#endif
+
 #ifdef HAVE_DTLS1_STOP_TIMER
 /* OpenSSL doesn't deliberately export this, but we need it to
    workaround a DTLS bug in versions < 1.0.0e */
 extern void dtls1_stop_timer(SSL *);
 #endif
 
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+/* Since OpenSSL 1.1, the SSL_SESSION structure is opaque and we can't
+ * just fill it in directly. So we have to generate the OpenSSL ASN.1
+ * representation of the SSL_SESSION, and use d2i_SSL_SESSION() to
+ * create the SSL_SESSION from that. */
+
+static void buf_append_INTEGER(struct oc_text_buf *buf, uint32_t datum)
+{
+	int l;
+
+	/* We only handle positive integers up to INT_MAX */
+	if (datum < 0x80)
+		l = 1;
+	else if (datum < 0x8000)
+		l = 2;
+	else if (datum < 0x800000)
+		l = 3;
+	else
+		l = 4;
+
+	if (buf_ensure_space(buf, 2 + l))
+		return;
+
+	buf->data[buf->pos++] = 0x02;
+	buf->data[buf->pos++] = l;
+	while (l--)
+		buf->data[buf->pos++] = datum >> (l * 8);
+}
+
+static void buf_append_OCTET_STRING(struct oc_text_buf *buf, void *data, int len)
+{
+	/* We only (need to) cope with length < 0x80 for now */
+	if (len >= 0x80) {
+		buf->error = -EINVAL;
+		return;
+	}
+
+	if (buf_ensure_space(buf, 2 + len))
+		return;
+
+	buf->data[buf->pos++] = 0x04;
+	buf->data[buf->pos++] = len;
+	memcpy(buf->data + buf->pos, data, len);
+	buf->pos += len;
+}
+
+static SSL_SESSION *generate_dtls_session(struct openconnect_info *vpninfo,
+					  int dtlsver, SSL_CIPHER *cipher)
+{
+	struct oc_text_buf *buf = buf_alloc();
+	SSL_SESSION *dtls_session;
+	const unsigned char *asn;
+	uint16_t cid;
+
+	buf_append_bytes(buf, "\x30\x80", 2); // SEQUENCE, indeterminate length
+	buf_append_INTEGER(buf, 1 /* SSL_SESSION_ASN1_VERSION */);
+	buf_append_INTEGER(buf, dtlsver);
+	store_be16(&cid, SSL_CIPHER_get_id(cipher) & 0xffff);
+	buf_append_OCTET_STRING(buf, &cid, 2);
+	buf_append_OCTET_STRING(buf, vpninfo->dtls_session_id,
+				sizeof(vpninfo->dtls_session_id));
+	buf_append_OCTET_STRING(buf, vpninfo->dtls_secret,
+				sizeof(vpninfo->dtls_secret));
+	/* If the length actually fits in one byte (which it should), do
+	 * it that way.  Else, leave it indeterminate and add two
+	 * end-of-contents octets to mark the end of the SEQUENCE. */
+	if (!buf_error(buf) && buf->pos <= 0x80)
+		buf->data[1] = buf->pos - 2;
+	else
+		buf_append_bytes(buf, "\0\0", 2);
+
+	if (buf_error(buf)) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to create SSL_SESSION ASN.1 for OpenSSL: %s\n"),
+			     strerror(buf_error(buf)));
+		buf_free(buf);
+		return NULL;
+	}
+
+	asn = (void *)buf->data;
+	dtls_session = d2i_SSL_SESSION(NULL, &asn, buf->pos);
+	buf_free(buf);
+	if (!dtls_session) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("OpenSSL failed to parse SSL_SESSION ASN.1\n"));
+		openconnect_report_ssl_errors(vpninfo);
+		return NULL;
+	}
+
+	return dtls_session;
+}
+#else /* OpenSSL before 1.1 */
+static SSL_SESSION *generate_dtls_session(struct openconnect_info *vpninfo,
+					  int dtlsver, SSL_CIPHER *cipher)
+{
+	SSL_SESSION *dtls_session = SSL_SESSION_new();
+	if (!dtls_session) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Initialise DTLSv1 session failed\n"));
+		return NULL;
+	}
+
+	dtls_session->ssl_version = dtlsver;
+	dtls_session->master_key_length = sizeof(vpninfo->dtls_secret);
+	memcpy(dtls_session->master_key, vpninfo->dtls_secret,
+	       sizeof(vpninfo->dtls_secret));
+
+	dtls_session->session_id_length = sizeof(vpninfo->dtls_session_id);
+	memcpy(dtls_session->session_id, vpninfo->dtls_session_id,
+	       sizeof(vpninfo->dtls_session_id));
+
+	dtls_session->cipher = cipher;
+	dtls_session->cipher_id = cipher->id;
+
+	return dtls_session;
+}
+#endif
+
 static int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 {
 	STACK_OF(SSL_CIPHER) *ciphers;
 	method_const SSL_METHOD *dtls_method;
-	SSL_CIPHER *dtls_cipher;
+	SSL_SESSION *dtls_session;
 	SSL *dtls_ssl;
 	BIO *dtls_bio;
+	int dtlsver = DTLS1_BAD_VER;
+	const char *cipher = vpninfo->dtls_cipher;
+
+#ifdef HAVE_DTLS12
+	if (!strcmp(cipher, "OC-DTLS1_2-AES128-GCM")) {
+		dtlsver = DTLS1_2_VERSION;
+		cipher = "AES128-GCM-SHA256";
+	} else if (!strcmp(cipher, "OC-DTLS1_2-AES256-GCM")) {
+		dtlsver = DTLS1_2_VERSION;
+		cipher = "AES256-GCM-SHA384";
+	}
+#endif
 
 	if (!vpninfo->dtls_ctx) {
-		dtls_method = DTLSv1_client_method();
+#ifdef HAVE_DTLS12
+		if (dtlsver == DTLS1_2_VERSION)
+			dtls_method = DTLSv1_2_client_method();
+		else
+#endif
+			dtls_method = DTLSv1_client_method();
+
 		vpninfo->dtls_ctx = SSL_CTX_new(dtls_method);
 		if (!vpninfo->dtls_ctx) {
 			vpn_progress(vpninfo, PRG_ERR,
@@ -132,7 +263,7 @@ static int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 		   away the tail of data packets. */
 		SSL_CTX_set_read_ahead(vpninfo->dtls_ctx, 1);
 
-		if (!SSL_CTX_set_cipher_list(vpninfo->dtls_ctx, vpninfo->dtls_cipher)) {
+		if (!SSL_CTX_set_cipher_list(vpninfo->dtls_ctx, cipher)) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Set DTLS cipher list failed\n"));
 			SSL_CTX_free(vpninfo->dtls_ctx);
@@ -142,27 +273,6 @@ static int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 		}
 	}
 
-	if (!vpninfo->dtls_session) {
-		/* We're going to "resume" a session which never existed. Fake it... */
-		vpninfo->dtls_session = SSL_SESSION_new();
-		if (!vpninfo->dtls_session) {
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Initialise DTLSv1 session failed\n"));
-			vpninfo->dtls_attempt_period = 0;
-			return -EINVAL;
-		}
-		vpninfo->dtls_session->ssl_version = 0x0100; /* DTLS1_BAD_VER */
-	}
-
-	/* Do this every time; it may have changed due to a rekey */
-	vpninfo->dtls_session->master_key_length = sizeof(vpninfo->dtls_secret);
-	memcpy(vpninfo->dtls_session->master_key, vpninfo->dtls_secret,
-	       sizeof(vpninfo->dtls_secret));
-
-	vpninfo->dtls_session->session_id_length = sizeof(vpninfo->dtls_session_id);
-	memcpy(vpninfo->dtls_session->session_id, vpninfo->dtls_session_id,
-	       sizeof(vpninfo->dtls_session_id));
-
 	dtls_ssl = SSL_new(vpninfo->dtls_ctx);
 	SSL_set_connect_state(dtls_ssl);
 
@@ -171,43 +281,54 @@ static int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 		vpn_progress(vpninfo, PRG_ERR, _("Not precisely one DTLS cipher\n"));
 		SSL_CTX_free(vpninfo->dtls_ctx);
 		SSL_free(dtls_ssl);
-		SSL_SESSION_free(vpninfo->dtls_session);
 		vpninfo->dtls_ctx = NULL;
-		vpninfo->dtls_session = NULL;
 		vpninfo->dtls_attempt_period = 0;
 		return -EINVAL;
 	}
-	dtls_cipher = sk_SSL_CIPHER_value(ciphers, 0);
 
-	/* Set the appropriate cipher on our session to be resumed */
-	vpninfo->dtls_session->cipher = dtls_cipher;
-	vpninfo->dtls_session->cipher_id = dtls_cipher->id;
+	/* We're going to "resume" a session which never existed. Fake it... */
+	dtls_session = generate_dtls_session(vpninfo, dtlsver,
+					     sk_SSL_CIPHER_value(ciphers, 0));
+	if (!dtls_session) {
+		SSL_CTX_free(vpninfo->dtls_ctx);
+		SSL_free(dtls_ssl);
+		vpninfo->dtls_ctx = NULL;
+		vpninfo->dtls_attempt_period = 0;
+		return -EINVAL;
+	}
 
 	/* Add the generated session to the SSL */
-	if (!SSL_set_session(dtls_ssl, vpninfo->dtls_session)) {
+	if (!SSL_set_session(dtls_ssl, dtls_session)) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("SSL_set_session() failed with old protocol version 0x%x\n"
 			       "Are you using a version of OpenSSL older than 0.9.8m?\n"
 			       "See http://rt.openssl.org/Ticket/Display.html?id=1751\n"
 			       "Use the --no-dtls command line option to avoid this message\n"),
-			     vpninfo->dtls_session->ssl_version);
+			     DTLS1_BAD_VER);
+		SSL_CTX_free(vpninfo->dtls_ctx);
+		SSL_free(dtls_ssl);
+		vpninfo->dtls_ctx = NULL;
 		vpninfo->dtls_attempt_period = 0;
+		SSL_SESSION_free(dtls_session);
 		return -EINVAL;
 	}
+	/* We don't need our own refcount on it any more */
+	SSL_SESSION_free(dtls_session);
 
 	dtls_bio = BIO_new_socket(dtls_fd, BIO_NOCLOSE);
 	/* Set non-blocking */
 	BIO_set_nbio(dtls_bio, 1);
 	SSL_set_bio(dtls_ssl, dtls_bio, dtls_bio);
 
-	SSL_set_options(dtls_ssl, SSL_OP_CISCO_ANYCONNECT);
+	if (dtlsver == DTLS1_BAD_VER)
+		SSL_set_options(dtls_ssl, SSL_OP_CISCO_ANYCONNECT);
 
 	vpninfo->dtls_ssl = dtls_ssl;
 
 	return 0;
 }
 
-int dtls_try_handshake(struct openconnect_info *vpninfo)
+static int dtls_try_handshake(struct openconnect_info *vpninfo)
 {
 	int ret = SSL_do_handshake(vpninfo->dtls_ssl);
 
@@ -311,7 +432,6 @@ void dtls_shutdown(struct openconnect_info *vpninfo)
 {
 	dtls_close(vpninfo);
 	SSL_CTX_free(vpninfo->dtls_ctx);
-	SSL_SESSION_free(vpninfo->dtls_session);
 }
 
 #elif defined(DTLS_GNUTLS)
@@ -398,7 +518,7 @@ static int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 	return 0;
 }
 
-int dtls_try_handshake(struct openconnect_info *vpninfo)
+static int dtls_try_handshake(struct openconnect_info *vpninfo)
 {
 	int err = gnutls_handshake(vpninfo->dtls_ssl);
 	char *str;
@@ -468,9 +588,9 @@ void dtls_shutdown(struct openconnect_info *vpninfo)
 }
 #endif
 
-int connect_dtls_socket(struct openconnect_info *vpninfo)
+static int connect_dtls_socket(struct openconnect_info *vpninfo)
 {
-	int dtls_fd, ret, sndbuf;
+	int dtls_fd, ret;
 
 	/* Sanity check for the removal of new_dtls_{fd,ssl} */
 	if (vpninfo->dtls_fd != -1) {
@@ -499,61 +619,10 @@ int connect_dtls_socket(struct openconnect_info *vpninfo)
 		return -EINVAL;
 	}
 
-	dtls_fd = socket(vpninfo->peer_addr->sa_family, SOCK_DGRAM, IPPROTO_UDP);
-	if (dtls_fd < 0) {
-		vpn_perror(vpninfo, _("Open UDP socket for DTLS:"));
+	dtls_fd = udp_connect(vpninfo);
+	if (dtls_fd < 0)
 		return -EINVAL;
-	}
-	if (vpninfo->protect_socket)
-		vpninfo->protect_socket(vpninfo->cbdata, dtls_fd);
 
-	sndbuf = vpninfo->ip_info.mtu * 2;
-	setsockopt(dtls_fd, SOL_SOCKET, SO_SNDBUF, (void *)&sndbuf, sizeof(sndbuf));
-
-	if (vpninfo->dtls_local_port) {
-		union {
-			struct sockaddr_in in;
-			struct sockaddr_in6 in6;
-		} dtls_bind_addr;
-		int dtls_bind_addrlen;
-		memset(&dtls_bind_addr, 0, sizeof(dtls_bind_addr));
-
-		if (vpninfo->peer_addr->sa_family == AF_INET) {
-			struct sockaddr_in *addr = &dtls_bind_addr.in;
-			dtls_bind_addrlen = sizeof(*addr);
-			addr->sin_family = AF_INET;
-			addr->sin_addr.s_addr = INADDR_ANY;
-			addr->sin_port = htons(vpninfo->dtls_local_port);
-		} else if (vpninfo->peer_addr->sa_family == AF_INET6) {
-			struct sockaddr_in6 *addr = &dtls_bind_addr.in6;
-			dtls_bind_addrlen = sizeof(*addr);
-			addr->sin6_family = AF_INET6;
-			addr->sin6_addr = in6addr_any;
-			addr->sin6_port = htons(vpninfo->dtls_local_port);
-		} else {
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Unknown protocol family %d. Cannot do DTLS\n"),
-				     vpninfo->peer_addr->sa_family);
-			vpninfo->dtls_attempt_period = 0;
-			closesocket(dtls_fd);
-			return -EINVAL;
-		}
-
-		if (bind(dtls_fd, (struct sockaddr *)&dtls_bind_addr, dtls_bind_addrlen)) {
-			vpn_perror(vpninfo, _("Bind UDP socket for DTLS"));
-			closesocket(dtls_fd);
-			return -EINVAL;
-		}
-	}
-
-	if (connect(dtls_fd, vpninfo->dtls_addr, vpninfo->peer_addrlen)) {
-		vpn_perror(vpninfo, _("UDP (DTLS) connect:\n"));
-		closesocket(dtls_fd);
-		return -EINVAL;
-	}
-
-	set_fd_cloexec(dtls_fd);
-	set_sock_nonblock(dtls_fd);
 
 	ret = start_dtls_handshake(vpninfo, dtls_fd);
 	if (ret) {
@@ -593,11 +662,14 @@ static int dtls_reconnect(struct openconnect_info *vpninfo)
 	return connect_dtls_socket(vpninfo);
 }
 
-int openconnect_setup_dtls(struct openconnect_info *vpninfo, int dtls_attempt_period)
+int dtls_setup(struct openconnect_info *vpninfo, int dtls_attempt_period)
 {
 	struct oc_vpn_option *dtls_opt = vpninfo->dtls_options;
 	int dtls_port = 0;
 
+	if (vpninfo->dtls_state == DTLS_DISABLED)
+		return -EINVAL;
+
 	vpninfo->dtls_attempt_period = dtls_attempt_period;
 	if (!dtls_attempt_period)
 		return 0;
@@ -646,27 +718,10 @@ int openconnect_setup_dtls(struct openconnect_info *vpninfo, int dtls_attempt_pe
 	if (vpninfo->dtls_times.rekey <= 0)
 		vpninfo->dtls_times.rekey_method = REKEY_NONE;
 
-	vpninfo->dtls_addr = malloc(vpninfo->peer_addrlen);
-	if (!vpninfo->dtls_addr) {
-		vpninfo->dtls_attempt_period = 0;
-		return -ENOMEM;
-	}
-	memcpy(vpninfo->dtls_addr, vpninfo->peer_addr, vpninfo->peer_addrlen);
-
-	if (vpninfo->peer_addr->sa_family == AF_INET) {
-		struct sockaddr_in *sin = (void *)vpninfo->dtls_addr;
-		sin->sin_port = htons(dtls_port);
-	} else if (vpninfo->peer_addr->sa_family == AF_INET6) {
-		struct sockaddr_in6 *sin = (void *)vpninfo->dtls_addr;
-		sin->sin6_port = htons(dtls_port);
-	} else {
-		vpn_progress(vpninfo, PRG_ERR,
-			     _("Unknown protocol family %d. Cannot do DTLS\n"),
-			     vpninfo->peer_addr->sa_family);
+	if (udp_sockaddr(vpninfo, dtls_port)) {
 		vpninfo->dtls_attempt_period = 0;
 		return -EINVAL;
 	}
-
 	if (connect_dtls_socket(vpninfo))
 		return -EINVAL;
 
@@ -754,6 +809,15 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 			vpn_progress(vpninfo, PRG_DEBUG, _("Got DTLS Keepalive\n"));
 			break;
 
+		case AC_PKT_COMPRESSED:
+			if (!vpninfo->dtls_compr) {
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Compressed DTLS packet received when compression not enabled\n"));
+				goto unknown_pkt;
+			}
+			decompress_and_queue_packet(vpninfo, vpninfo->dtls_compr,
+						    vpninfo->dtls_pkt->data, len - 1);
+			break;
 		default:
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Unknown DTLS packet type %02x, len %d\n"),
@@ -766,6 +830,7 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 				 * the appropriate length of garbage. So don't abort... for now. */
 				break;
 			} else {
+			unknown_pkt:
 				vpninfo->quit_reason = "Unknown packet received";
 				return 1;
 			}
@@ -814,7 +879,7 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	case KA_KEEPALIVE:
 		/* No need to send an explicit keepalive
 		   if we have real data to send */
-		if (vpninfo->outgoing_queue)
+		if (vpninfo->outgoing_queue.head)
 			break;
 
 		vpn_progress(vpninfo, PRG_DEBUG, _("Send DTLS Keepalive\n"));
@@ -833,26 +898,32 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 
 	/* Service outgoing packet queue */
 	unmonitor_write_fd(vpninfo, dtls);
-	while (vpninfo->outgoing_queue) {
-		struct pkt *this = vpninfo->outgoing_queue;
+	while (vpninfo->outgoing_queue.head) {
+		struct pkt *this = dequeue_packet(&vpninfo->outgoing_queue);
+		struct pkt *send_pkt = this;
 		int ret;
 
-		vpninfo->outgoing_queue = this->next;
-		vpninfo->outgoing_qlen--;
-
 		/* One byte of header */
-		this->hdr[7] = AC_PKT_DATA;
+		this->cstp.hdr[7] = AC_PKT_DATA;
+
+		/* We can compress into vpninfo->deflate_pkt unless CSTP
+		 * currently has a compressed packet pending — which it
+		 * shouldn't if DTLS is active. */
+		if (vpninfo->dtls_compr &&
+		    vpninfo->current_ssl_pkt != vpninfo->deflate_pkt &&
+		    !compress_packet(vpninfo, vpninfo->dtls_compr, this)) {
+				send_pkt = vpninfo->deflate_pkt;
+				send_pkt->cstp.hdr[7] = AC_PKT_COMPRESSED;
+		}
 
 #if defined(DTLS_OPENSSL)
-		ret = SSL_write(vpninfo->dtls_ssl, &this->hdr[7], this->len + 1);
+		ret = SSL_write(vpninfo->dtls_ssl, &send_pkt->cstp.hdr[7], send_pkt->len + 1);
 		if (ret <= 0) {
 			ret = SSL_get_error(vpninfo->dtls_ssl, ret);
 
 			if (ret == SSL_ERROR_WANT_WRITE) {
 				monitor_write_fd(vpninfo, dtls);
-				vpninfo->outgoing_queue = this;
-				vpninfo->outgoing_qlen++;
-
+				requeue_packet(&vpninfo->outgoing_queue, this);
 			} else if (ret != SSL_ERROR_WANT_READ) {
 				/* If it's a real error, kill the DTLS connection and
 				   requeue the packet to be sent over SSL */
@@ -861,29 +932,26 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 					     ret);
 				openconnect_report_ssl_errors(vpninfo);
 				dtls_reconnect(vpninfo);
-				vpninfo->outgoing_queue = this;
-				vpninfo->outgoing_qlen++;
+				requeue_packet(&vpninfo->outgoing_queue, this);
 				work_done = 1;
 			}
 			return work_done;
 		}
 #elif defined(DTLS_GNUTLS)
-		ret = gnutls_record_send(vpninfo->dtls_ssl, &this->hdr[7], this->len + 1);
+		ret = gnutls_record_send(vpninfo->dtls_ssl, &send_pkt->cstp.hdr[7], send_pkt->len + 1);
 		if (ret <= 0) {
 			if (ret != GNUTLS_E_AGAIN) {
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("DTLS got write error: %s. Falling back to SSL\n"),
 					     gnutls_strerror(ret));
 				dtls_reconnect(vpninfo);
-				vpninfo->outgoing_queue = this;
-				vpninfo->outgoing_qlen++;
 				work_done = 1;
-			} else if (gnutls_record_get_direction(vpninfo->dtls_ssl)) {
+			} else {
+				/* Wake me up when it becomes writeable */
 				monitor_write_fd(vpninfo, dtls);
-				vpninfo->outgoing_queue = this;
-				vpninfo->outgoing_qlen++;
 			}
 
+			requeue_packet(&vpninfo->outgoing_queue, this);
 			return work_done;
 		}
 #endif
@@ -898,18 +966,4 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 }
 #else /* !HAVE_DTLS */
 #warning Your SSL library does not seem to support Cisco DTLS compatibility
-int openconnect_setup_dtls(struct openconnect_info *vpninfo, int dtls_attempt_period)
-{
-	vpn_progress(vpninfo, PRG_ERR,
-		     _("Built against SSL library with no Cisco DTLS support\n"));
-	return -EINVAL;
-}
-
-void dtls_close(struct openconnect_info *vpninfo)
-{
-}
-
-void dtls_shutdown(struct openconnect_info *vpninfo)
-{
-}
 #endif
