@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2014 Intel Corporation.
+ * Copyright © 2008-2015 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -214,6 +214,57 @@ static int openconnect_gnutls_gets(struct openconnect_info *vpninfo, char *buf,
 	return i ?: ret;
 }
 
+int ssl_nonblock_read(struct openconnect_info *vpninfo, void *buf, int maxlen)
+{
+	int ret;
+
+	ret = gnutls_record_recv(vpninfo->https_sess, buf, maxlen);
+	if (ret > 0)
+		return ret;
+
+	if (ret != GNUTLS_E_AGAIN) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("SSL read error: %s; reconnecting.\n"),
+			     gnutls_strerror(ret));
+		return -EIO;
+	}
+	return 0;
+}
+
+int ssl_nonblock_write(struct openconnect_info *vpninfo, void *buf, int buflen)
+{
+	int ret;
+
+	ret = gnutls_record_send(vpninfo->https_sess, buf, buflen);
+	if (ret > 0)
+		return ret;
+
+	if (ret == GNUTLS_E_AGAIN) {
+		/*
+		 * Before 3.3.13, GnuTLS could return zero instead of one,
+		 * indicating that it was waiting for a read when in fact
+		 * it was waiting for a write. That caused us to block for
+		 * ever, waiting for the read that it said it wanted.
+		 *
+		 * So instead, just *assume* it actually wants a write.
+		 * Which is true most of the time, and on the rare occasion
+		 * that it *isn't* true, the failure mode will just be that
+		 * we keep waking up and calling GnuTLS again until the read
+		 * that it's waiting for does arrive.
+		 */
+		if (GNUTLS_VERSION_NUMBER < 0x03030d ||
+		    gnutls_record_get_direction(vpninfo->https_sess)) {
+			/* Waiting for the socket to become writable — it's
+			   probably stalled, and/or the buffers are full */
+			monitor_write_fd(vpninfo, ssl);
+		}
+		return 0;
+	}
+	vpn_progress(vpninfo, PRG_ERR, _("SSL send failed: %s\n"),
+		     gnutls_strerror(ret));
+	return -1;
+}
+
 static int check_certificate_expiry(struct openconnect_info *vpninfo, gnutls_x509_crt_t cert)
 {
 	const char *reason = NULL;
@@ -2494,14 +2545,15 @@ static P11KitPin *p11kit_pin_callback(const char *pin_source, P11KitUri *pin_uri
 
 #ifdef HAVE_LIBPCSCLITE
 int openconnect_hash_yubikey_password(struct openconnect_info *vpninfo,
-				      const char *password, const void *ident, int id_len)
+				      const char *password, int pwlen,
+				      const void *ident, int id_len)
 {
 	unsigned char U[SHA1_SIZE];
 	gnutls_hmac_hd_t dgst;
 	int ret = -EIO;
 	int i, j;
 
-	if (gnutls_hmac_init(&dgst, GNUTLS_MAC_SHA1, password, strlen(password)))
+	if (gnutls_hmac_init(&dgst, GNUTLS_MAC_SHA1, password, pwlen))
 		return -EIO;
 
 	if (gnutls_hmac(dgst, ident, id_len))
@@ -2540,3 +2592,43 @@ int openconnect_yubikey_chalresp(struct openconnect_info *vpninfo,
 	return 0;
 }
 #endif
+
+int hotp_hmac(struct openconnect_info *vpninfo, const void *challenge)
+{
+	int ret;
+	int hpos;
+	unsigned char hash[64]; /* Enough for a SHA256 */
+	gnutls_mac_algorithm_t alg;
+
+	switch(vpninfo->oath_hmac_alg) {
+	case OATH_ALG_HMAC_SHA1:
+		alg = GNUTLS_MAC_SHA1;
+		hpos = 19;
+		break;
+	case OATH_ALG_HMAC_SHA256:
+		alg = GNUTLS_MAC_SHA256;
+		hpos = 31;
+		break;
+	case OATH_ALG_HMAC_SHA512:
+		alg = GNUTLS_MAC_SHA512;
+		hpos = 63;
+		break;
+	default:
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Unsupported OATH HMAC algorithm\n"));
+		return -EINVAL;
+	}
+
+	ret = gnutls_hmac_fast(alg, vpninfo->oath_secret,
+			       vpninfo->oath_secret_len,
+			       challenge, 8, hash);
+	if (ret) {
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  _("Failed to calculate OATH HMAC: %s\n"),
+				  gnutls_strerror(ret));
+		return -EINVAL;
+	}
+
+	hpos = hash[hpos] & 15;
+	return load_be32(&hash[hpos]) & 0x7fffffff;
+}
