@@ -32,7 +32,6 @@ struct libctx {
 	jobject jobj;
 	jobject async_lock;
 	struct openconnect_info *vpninfo;
-	OPENCONNECT_X509 *cert;
 	int cmd_fd;
 	int loglevel;
 };
@@ -81,7 +80,7 @@ static jstring dup_to_jstring(JNIEnv *jenv, const char *in)
 	return in ? (*jenv)->NewStringUTF(jenv, in) : NULL;
 }
 
-static int dup_to_cstring(JNIEnv *jenv, jstring in, char **out)
+static int get_cstring(JNIEnv *jenv, jstring in, const char **out)
 {
 	const char *tmp;
 
@@ -96,16 +95,16 @@ static int dup_to_cstring(JNIEnv *jenv, jstring in, char **out)
 		return -1;
 	}
 
-	*out = strdup(tmp);
-	(*jenv)->ReleaseStringUTFChars(jenv, in, tmp);
-
-	if (!*out) {
-		OOM(jenv);
-		return -1;
-	}
+	*out = tmp;
 	return 0;
 }
 
+static void release_cstring(JNIEnv *jenv, jstring jstr, const char *cstr)
+{
+	if (cstr)
+		(*jenv)->ReleaseStringUTFChars(jenv, jstr, cstr);
+}
+
 static int set_int(struct libctx *ctx, jobject jobj, const char *name, int value)
 {
 	jclass jcls = (*ctx->jenv)->GetObjectClass(ctx->jenv, jobj);
@@ -195,7 +194,7 @@ static int add_string_pair(struct libctx *ctx, jclass jcls, jobject jobj,
 	return 0;
 }
 
-static int validate_peer_cert_cb(void *privdata, OPENCONNECT_X509 *cert, const char *reason)
+static int validate_peer_cert_cb(void *privdata, const char *reason)
 {
 	struct libctx *ctx = privdata;
 	jstring jreason;
@@ -209,7 +208,6 @@ static int validate_peer_cert_cb(void *privdata, OPENCONNECT_X509 *cert, const c
 	if (!jreason)
 		goto out;
 
-	ctx->cert = cert;
 	mid = get_obj_mid(ctx, ctx->jobj, "onValidatePeerCert", "(Ljava/lang/String;)I");
 	if (mid)
 		ret = (*ctx->jenv)->CallIntMethod(ctx->jenv, ctx->jobj, mid, jreason);
@@ -219,7 +217,7 @@ out:
 	return ret;
 }
 
-static int write_new_config_cb(void *privdata, char *buf, int buflen)
+static int write_new_config_cb(void *privdata, const char *buf, int buflen)
 {
 	struct libctx *ctx = privdata;
 	jmethodID mid;
@@ -391,7 +389,7 @@ static int add_form_option(struct libctx *ctx, jobject jform, struct oc_form_opt
 	if (set_int(ctx, jopt, "type", opt->type) ||
 	    set_string(ctx, jopt, "name", opt->name) ||
 	    set_string(ctx, jopt, "label", opt->label) ||
-	    set_string(ctx, jopt, "value", opt->value) ||
+	    set_string(ctx, jopt, "value", opt->_value) ||
 	    set_long(ctx, jopt, "flags", opt->flags))
 		return -1;
 
@@ -406,7 +404,7 @@ static char *lookup_choice_name(struct oc_form_opt_select *opt, const char *name
 {
 	int i;
 
-	/* opt->value is NOT a caller-allocated string for OC_FORM_OPT_SELECT */
+	/* opt->_value is NOT a caller-allocated string for OC_FORM_OPT_SELECT */
 	for (i = 0; i < opt->nr_choices; i++)
 		if (!strcmp(opt->choices[i]->name, name))
 			return opt->choices[i]->name;
@@ -465,11 +463,11 @@ static int process_auth_form_cb(void *privdata, struct oc_auth_form *form)
 				goto err;
 
 			if (opt->type == OC_FORM_OPT_SELECT)
-				opt->value = lookup_choice_name((void *)opt, tmp);
+				opt->_value = lookup_choice_name((void *)opt, tmp);
 			else {
-				free(opt->value);
-				opt->value = strdup(tmp);
-				if (!opt->value)
+				free(opt->_value);
+				opt->_value = strdup(tmp);
+				if (!opt->_value)
 					OOM(ctx->jenv);
 			}
 			(*ctx->jenv)->ReleaseStringUTFChars(ctx->jenv, jvalue, tmp);
@@ -484,7 +482,8 @@ err:
 	return -1;
 }
 
-static void progress_cb(void *privdata, int level, const char *fmt, ...)
+static void __attribute__ ((format(printf, 3, 4)))
+    progress_cb(void *privdata, int level, const char *fmt, ...)
 {
 	struct libctx *ctx = privdata;
 	va_list ap;
@@ -525,6 +524,46 @@ out:
 	(*ctx->jenv)->PopLocalFrame(ctx->jenv, NULL);
 }
 
+static int lock_token_cb(void *privdata)
+{
+	struct libctx *ctx = privdata;
+	jmethodID mid;
+	int ret = -1;
+
+	if ((*ctx->jenv)->PushLocalFrame(ctx->jenv, 256) < 0)
+		return -1;
+
+	mid = get_obj_mid(ctx, ctx->jobj, "onTokenLock", "()I");
+	if (mid)
+		ret = (*ctx->jenv)->CallIntMethod(ctx->jenv, ctx->jobj, mid);
+
+	(*ctx->jenv)->PopLocalFrame(ctx->jenv, NULL);
+	return ret;
+}
+
+static int unlock_token_cb(void *privdata, const char *new_token)
+{
+	struct libctx *ctx = privdata;
+	jstring jtoken;
+	int ret = -1;
+	jmethodID mid;
+
+	if ((*ctx->jenv)->PushLocalFrame(ctx->jenv, 256) < 0)
+		return -1;
+
+	jtoken = dup_to_jstring(ctx->jenv, new_token);
+	if (!jtoken)
+		goto out;
+
+	mid = get_obj_mid(ctx, ctx->jobj, "onTokenUnlock", "(Ljava/lang/String;)I");
+	if (mid)
+		ret = (*ctx->jenv)->CallIntMethod(ctx->jenv, ctx->jobj, mid, jtoken);
+
+out:
+	(*ctx->jenv)->PopLocalFrame(ctx->jenv, NULL);
+	return ret;
+}
+
 /* Library init/uninit */
 
 static jobject init_async_lock(struct libctx *ctx)
@@ -566,6 +605,8 @@ JNIEXPORT jlong JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_init(
 	if (!ctx->vpninfo)
 		goto bad_delete_ref;
 
+	openconnect_set_token_callbacks(ctx->vpninfo, ctx, lock_token_cb,
+					unlock_token_cb);
 	openconnect_set_protect_socket_handler(ctx->vpninfo, protect_socket_cb);
 	openconnect_set_stats_handler(ctx->vpninfo, stats_cb);
 
@@ -642,46 +683,42 @@ JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_obtainCo
 	JNIEnv *jenv, jobject jobj)
 {
 	struct libctx *ctx = getctx(jenv, jobj);
-	int ret;
 
 	if (!ctx)
 		return 0;
-	ctx->cert = NULL;
-	ret = openconnect_obtain_cookie(ctx->vpninfo);
-	if (ret == 0)
-		ctx->cert = openconnect_get_peer_cert(ctx->vpninfo);
-	return ret;
+	return openconnect_obtain_cookie(ctx->vpninfo);
 }
 
 /* special handling: caller-allocated buffer */
-JNIEXPORT jstring JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getCertSHA1(
+JNIEXPORT jstring JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getPeerCertHash(
 	JNIEnv *jenv, jobject jobj)
 {
 	struct libctx *ctx = getctx(jenv, jobj);
-	char buf[41];
+	const char *hash;
 	jstring jresult = NULL;
 
-	if (!ctx || !ctx->cert)
+	if (!ctx)
 		return NULL;
-	if (openconnect_get_cert_sha1(ctx->vpninfo, ctx->cert, buf))
+	hash = openconnect_get_peer_cert_hash(ctx->vpninfo);
+	if (!hash)
 		return NULL;
-	jresult = dup_to_jstring(ctx->jenv, buf);
+	jresult = dup_to_jstring(ctx->jenv, hash);
 	if (!jresult)
 		OOM(ctx->jenv);
 	return jresult;
 }
 
 /* special handling: callee-allocated, caller-freed string */
-JNIEXPORT jstring JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getCertDetails(
+JNIEXPORT jstring JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getPeerCertDetails(
 	JNIEnv *jenv, jobject jobj)
 {
 	struct libctx *ctx = getctx(jenv, jobj);
 	char *buf = NULL;
 	jstring jresult = NULL;
 
-	if (!ctx || !ctx->cert)
+	if (!ctx)
 		return NULL;
-	buf = openconnect_get_cert_details(ctx->vpninfo, ctx->cert);
+	buf = openconnect_get_peer_cert_details(ctx->vpninfo);
 	if (!buf)
 		return NULL;
 
@@ -689,12 +726,12 @@ JNIEXPORT jstring JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getCe
 	if (!jresult)
 		OOM(ctx->jenv);
 
-	free(buf);
+	openconnect_free_cert_info(ctx->vpninfo, buf);
 	return jresult;
 }
 
 /* special handling: callee-allocated, caller-freed binary buffer */
-JNIEXPORT jbyteArray JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getCertDER(
+JNIEXPORT jbyteArray JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getPeerCertDER(
 	JNIEnv *jenv, jobject jobj)
 {
 	struct libctx *ctx = getctx(jenv, jobj);
@@ -702,9 +739,9 @@ JNIEXPORT jbyteArray JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_ge
 	int ret;
 	jbyteArray jresult = NULL;
 
-	if (!ctx || !ctx->cert)
+	if (!ctx)
 		return NULL;
-	ret = openconnect_get_cert_DER(ctx->vpninfo, ctx->cert, &buf);
+	ret = openconnect_get_peer_cert_DER(ctx->vpninfo, &buf);
 	if (ret < 0)
 		return NULL;
 
@@ -712,7 +749,7 @@ JNIEXPORT jbyteArray JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_ge
 	if (jresult)
 		(*ctx->jenv)->SetByteArrayRegion(ctx->jenv, jresult, 0, ret, (jbyte *) buf);
 
-	free(buf);
+	openconnect_free_cert_info(ctx->vpninfo, buf);
 	return jresult;
 }
 
@@ -721,17 +758,15 @@ JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setClien
 	JNIEnv *jenv, jobject jobj, jstring jcert, jstring jsslkey)
 {
 	struct libctx *ctx = getctx(jenv, jobj);
-	char *cert = NULL, *sslkey = NULL;
+	const char *cert = NULL, *sslkey = NULL;
 
-	if (!ctx ||
-	    dup_to_cstring(ctx->jenv, jcert, &cert) ||
-	    dup_to_cstring(ctx->jenv, jsslkey, &sslkey)) {
-		free(cert);
-		free(sslkey);
-		return;
-	}
+	if (ctx && !get_cstring(ctx->jenv, jcert, &cert) &&
+	    !get_cstring(ctx->jenv, jsslkey, &sslkey))
+		openconnect_set_client_cert(ctx->vpninfo, cert, sslkey);
 
-	openconnect_set_client_cert(ctx->vpninfo, cert, sslkey);
+	release_cstring(ctx->jenv, jcert, cert);
+	release_cstring(ctx->jenv, jsslkey, sslkey);
+	return;
 }
 
 /* special handling: multiple string arguments */
@@ -739,60 +774,54 @@ JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setupTun
 	JNIEnv *jenv, jobject jobj, jstring jarg0, jstring jarg1)
 {
 	struct libctx *ctx = getctx(jenv, jobj);
-	char *arg0 = NULL, *arg1 = NULL;
+	const char *arg0 = NULL, *arg1 = NULL;
+	int ret = -ENOMEM;
 
-	if (!ctx ||
-	    dup_to_cstring(ctx->jenv, jarg0, &arg0) ||
-	    dup_to_cstring(ctx->jenv, jarg1, &arg1)) {
-		free(arg0);
-		free(arg1);
-		return -ENOMEM;
-	}
-	return openconnect_setup_tun_device(ctx->vpninfo, arg0, arg1);
+	if (ctx && !get_cstring(ctx->jenv, jarg0, &arg0) &&
+	    !get_cstring(ctx->jenv, jarg1, &arg1))
+		ret = openconnect_setup_tun_device(ctx->vpninfo, arg0, arg1);
+
+	release_cstring(ctx->jenv, jarg0, arg0);
+	release_cstring(ctx->jenv, jarg1, arg1);
+	return ret;
 }
 
 JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setCSDWrapper(
 	JNIEnv *jenv, jobject jobj, jstring jarg0, jstring jarg1, jstring jarg2)
 {
 	struct libctx *ctx = getctx(jenv, jobj);
-	char *arg0 = NULL, *arg1 = NULL, *arg2 = NULL;
+	const char *arg0 = NULL, *arg1 = NULL, *arg2 = NULL;
 
-	if (!ctx ||
-	    dup_to_cstring(ctx->jenv, jarg0, &arg0) ||
-	    dup_to_cstring(ctx->jenv, jarg1, &arg1) ||
-	    dup_to_cstring(ctx->jenv, jarg2, &arg2)) {
-		free(arg0);
-		free(arg1);
-		free(arg2);
-		return;
+	if (ctx &&
+	    !get_cstring(ctx->jenv, jarg0, &arg0) &&
+	    !get_cstring(ctx->jenv, jarg1, &arg1) &&
+	    !get_cstring(ctx->jenv, jarg2, &arg2)) {
+		openconnect_setup_csd(ctx->vpninfo, getuid(), 1, arg0);
+
+		openconnect_set_csd_environ(ctx->vpninfo, "TMPDIR", arg1);
+		openconnect_set_csd_environ(ctx->vpninfo, "PATH", arg2);
 	}
-	openconnect_setup_csd(ctx->vpninfo, getuid(), 1, arg0);
 
-	if (arg1)
-		setenv("TMPDIR", arg1, 1);
-	free(arg1);
-
-	if (arg2)
-		setenv("PATH", arg2, 1);
-	free(arg2);
+	release_cstring(ctx->jenv, jarg0, arg0);
+	release_cstring(ctx->jenv, jarg1, arg1);
+	release_cstring(ctx->jenv, jarg2, arg2);
 }
 
 JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setMobileInfo(
 	JNIEnv *jenv, jobject jobj, jstring jarg0, jstring jarg1, jstring jarg2)
 {
 	struct libctx *ctx = getctx(jenv, jobj);
-	char *arg0 = NULL, *arg1 = NULL, *arg2 = NULL;
+	const char *arg0 = NULL, *arg1 = NULL, *arg2 = NULL;
 
-	if (!ctx ||
-	    dup_to_cstring(ctx->jenv, jarg0, &arg0) ||
-	    dup_to_cstring(ctx->jenv, jarg1, &arg1) ||
-	    dup_to_cstring(ctx->jenv, jarg2, &arg2)) {
-		free(arg0);
-		free(arg1);
-		free(arg2);
-		return;
-	}
-	openconnect_set_mobile_info(ctx->vpninfo, arg0, arg1, arg2);
+	if (ctx &&
+	    !get_cstring(ctx->jenv, jarg0, &arg0) &&
+	    !get_cstring(ctx->jenv, jarg1, &arg1) &&
+	    !get_cstring(ctx->jenv, jarg2, &arg2))
+		openconnect_set_mobile_info(ctx->vpninfo, arg0, arg1, arg2);
+
+	release_cstring(ctx->jenv, jarg0, arg0);
+	release_cstring(ctx->jenv, jarg1, arg1);
+	release_cstring(ctx->jenv, jarg2, arg2);
 }
 
 /* class methods (general library info) */
@@ -827,6 +856,12 @@ JNIEXPORT jboolean JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_hasO
 	return openconnect_has_oath_support();
 }
 
+JNIEXPORT jboolean JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_hasYubiOATHSupport(
+	JNIEnv *jenv, jclass jcls)
+{
+	return openconnect_has_yubioath_support();
+}
+
 /* simple cases: void or int params */
 
 JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getPort(
@@ -909,6 +944,16 @@ JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setPFS(
 	openconnect_set_pfs(ctx->vpninfo, arg);
 }
 
+JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setSystemTrust(
+	JNIEnv *jenv, jobject jobj, jboolean arg)
+{
+	struct libctx *ctx = getctx(jenv, jobj);
+
+	if (!ctx)
+		return;
+	openconnect_set_system_trust(ctx->vpninfo, arg);
+}
+
 JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_makeCSTPConnection(
 	JNIEnv *jenv, jobject jobj)
 {
@@ -1021,18 +1066,49 @@ JNIEXPORT jstring JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getIF
 	RETURN_STRING_END
 }
 
+JNIEXPORT jstring JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getDTLSCipher(
+	JNIEnv *jenv, jobject jobj)
+{
+	RETURN_STRING_START
+	buf = openconnect_get_dtls_cipher(ctx->vpninfo);
+	RETURN_STRING_END
+}
+
+JNIEXPORT jstring JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getCSTPCipher(
+	JNIEnv *jenv, jobject jobj)
+{
+	RETURN_STRING_START
+	buf = openconnect_get_cstp_cipher(ctx->vpninfo);
+	RETURN_STRING_END
+}
+
 #define SET_STRING_START(ret) \
 	struct libctx *ctx = getctx(jenv, jobj); \
-	char *arg; \
-	if (dup_to_cstring(ctx->jenv, jarg, &arg)) \
+	const char *arg = NULL;			 \
+	if (get_cstring(ctx->jenv, jarg, &arg)) \
 		return ret;
 
+#define SET_STRING_END() \
+	release_cstring(ctx->jenv, jarg, arg)
+
+JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_checkPeerCertHash(
+	JNIEnv *jenv, jobject jobj, jstring jarg)
+{
+	int ret;
+	SET_STRING_START(-ENOMEM)
+	ret = openconnect_check_peer_cert_hash(ctx->vpninfo, arg);
+	SET_STRING_END();
+
+	return ret;
+}
+
 JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_parseURL(
 	JNIEnv *jenv, jobject jobj, jstring jarg)
 {
 	int ret;
 	SET_STRING_START(-ENOMEM)
 	ret = openconnect_parse_url(ctx->vpninfo, arg);
+	SET_STRING_END();
 	return ret;
 }
 
@@ -1042,6 +1118,7 @@ JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setProxy
 	int ret;
 	SET_STRING_START(-ENOMEM)
 	ret = openconnect_set_proxy_auth(ctx->vpninfo, arg);
+	SET_STRING_END();
 	return ret;
 }
 
@@ -1051,6 +1128,7 @@ JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setHTTPP
 	int ret;
 	SET_STRING_START(-ENOMEM)
 	ret = openconnect_set_http_proxy(ctx->vpninfo, arg);
+	SET_STRING_END();
 	return ret;
 }
 
@@ -1059,7 +1137,7 @@ JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setXMLSH
 {
 	SET_STRING_START()
 	openconnect_set_xmlsha1(ctx->vpninfo, arg, strlen(arg) + 1);
-	free(arg);
+	SET_STRING_END();
 }
 
 JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setHostname(
@@ -1067,6 +1145,7 @@ JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setHostn
 {
 	SET_STRING_START()
 	openconnect_set_hostname(ctx->vpninfo, arg);
+	SET_STRING_END();
 }
 
 JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setUrlpath(
@@ -1074,6 +1153,7 @@ JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setUrlpa
 {
 	SET_STRING_START()
 	openconnect_set_urlpath(ctx->vpninfo, arg);
+	SET_STRING_END();
 }
 
 JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setCAFile(
@@ -1081,6 +1161,7 @@ JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setCAFil
 {
 	SET_STRING_START()
 	openconnect_set_cafile(ctx->vpninfo, arg);
+	SET_STRING_END();
 }
 
 JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setReportedOS(
@@ -1088,6 +1169,7 @@ JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setRepor
 {
 	SET_STRING_START(-ENOMEM)
 	return openconnect_set_reported_os(ctx->vpninfo, arg);
+	SET_STRING_END();
 }
 
 JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setTokenMode(
@@ -1096,7 +1178,7 @@ JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setToken
 	int ret;
 	SET_STRING_START(-ENOMEM)
 	ret = openconnect_set_token_mode(ctx->vpninfo, mode, arg);
-	free(arg);
+	SET_STRING_END();
 	return ret;
 }
 
@@ -1106,16 +1188,10 @@ JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setupTun
 	int ret;
 	SET_STRING_START(-ENOMEM)
 	ret = openconnect_setup_tun_script(ctx->vpninfo, arg);
+	SET_STRING_END();
 	return ret;
 }
 
-JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setServerCertSHA1(
-	JNIEnv *jenv, jobject jobj, jstring jarg)
-{
-	SET_STRING_START()
-	openconnect_set_server_cert_sha1(ctx->vpninfo, arg);
-}
-
 JNIEXPORT jobject JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getIPInfo(
 	JNIEnv *jenv, jobject jobj)
 {
