@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2014 Intel Corporation.
+ * Copyright © 2008-2015 Intel Corporation.
  * Copyright © 2013 John Morrissey <jwm@horde.net>
  *
  * Authors: David Woodhouse <dwmw2@infradead.org>
@@ -29,10 +29,6 @@
 #include <stoken.h>
 #endif
 
-#ifdef HAVE_LIBOATH
-#include <liboath/oath.h>
-#endif
-
 #include <libxml/tree.h>
 #include <zlib.h>
 
@@ -69,10 +65,14 @@ struct openconnect_info *openconnect_vpninfo_new(const char *useragent,
 #ifndef _WIN32
 	vpninfo->tun_fd = -1;
 #endif
+	init_pkt_queue(&vpninfo->incoming_queue);
+	init_pkt_queue(&vpninfo->outgoing_queue);
+	init_pkt_queue(&vpninfo->oncp_control_queue);
 	vpninfo->ssl_fd = vpninfo->dtls_fd = -1;
 	vpninfo->cmd_fd = vpninfo->cmd_fd_write = -1;
+	vpninfo->tncc_fd = -1;
 	vpninfo->cert_expire_warning = 60 * 86400;
-	vpninfo->req_compr = COMPR_ALL;
+	vpninfo->req_compr = COMPR_STATELESS;
 	vpninfo->max_qlen = 10;
 	vpninfo->localname = strdup("localhost");
 	vpninfo->useragent = openconnect_create_useragent(useragent);
@@ -82,6 +82,10 @@ struct openconnect_info *openconnect_vpninfo_new(const char *useragent,
 	vpninfo->progress = progress;
 	vpninfo->cbdata = privdata ? : vpninfo;
 	vpninfo->xmlpost = 1;
+	vpninfo->verbose = PRG_TRACE;
+	vpninfo->try_http_auth = 1;
+	vpninfo->proxy_auth[AUTH_TYPE_BASIC].state = AUTH_DEFAULT_DISABLED;
+	vpninfo->http_auth[AUTH_TYPE_BASIC].state = AUTH_DEFAULT_DISABLED;
 	openconnect_set_reported_os(vpninfo, NULL);
 
 	if (!vpninfo->localname || !vpninfo->useragent)
@@ -90,7 +94,7 @@ struct openconnect_info *openconnect_vpninfo_new(const char *useragent,
 #ifdef ENABLE_NLS
 	bindtextdomain("openconnect", LOCALEDIR);
 #endif
-
+	openconnect_set_protocol(vpninfo, "anyconnect");
 	return vpninfo;
 
 err:
@@ -100,6 +104,73 @@ err:
 	return NULL;
 }
 
+int openconnect_set_protocol(struct openconnect_info *vpninfo, const char *protocol)
+{
+	if (!strcmp(protocol, "anyconnect")) {
+		vpninfo->proto.vpn_close_session = cstp_bye;
+		vpninfo->proto.tcp_connect = cstp_connect;
+		vpninfo->proto.tcp_mainloop = cstp_mainloop;
+		vpninfo->proto.add_http_headers = cstp_common_headers;
+		vpninfo->proto.obtain_cookie = cstp_obtain_cookie;
+#ifdef HAVE_DTLS
+		vpninfo->proto.udp_setup = dtls_setup;
+		vpninfo->proto.udp_mainloop = dtls_mainloop;
+		vpninfo->proto.udp_close = dtls_close;
+		vpninfo->proto.udp_shutdown = dtls_shutdown;
+#else
+		vpninfo->dtls_state = DTLS_DISABLED;
+#endif
+	} else if (!strcmp(protocol, "nc")) {
+		vpninfo->proto.vpn_close_session = NULL;
+		vpninfo->proto.tcp_connect = oncp_connect;
+		vpninfo->proto.tcp_mainloop = oncp_mainloop;
+		vpninfo->proto.add_http_headers = oncp_common_headers;
+		vpninfo->proto.obtain_cookie = oncp_obtain_cookie;
+#if defined(ESP_GNUTLS) || defined(ESP_OPENSSL)
+		vpninfo->proto.udp_setup = esp_setup;
+		vpninfo->proto.udp_mainloop = esp_mainloop;
+		vpninfo->proto.udp_close = esp_close;
+		vpninfo->proto.udp_shutdown = esp_shutdown;
+#else
+		vpninfo->dtls_state = DTLS_DISABLED;
+#endif
+	} else {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Unknown VPN protocol '%s'\n"), protocol);
+		return -EINVAL;
+	}
+
+	return 0;
+}
+
+void openconnect_set_loglevel(struct openconnect_info *vpninfo,
+			      int level)
+{
+	vpninfo->verbose = level;
+}
+
+int openconnect_setup_dtls(struct openconnect_info *vpninfo,
+			   int attempt_period)
+
+{
+	if (vpninfo->proto.udp_setup)
+		return vpninfo->proto.udp_setup(vpninfo, attempt_period);
+
+	vpn_progress(vpninfo, PRG_ERR,
+		     _("Built against SSL library with no Cisco DTLS support\n"));
+	return -EINVAL;
+}
+
+int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
+{
+	return vpninfo->proto.obtain_cookie(vpninfo);
+}
+
+int openconnect_make_cstp_connection(struct openconnect_info *vpninfo)
+{
+	return vpninfo->proto.tcp_connect(vpninfo);
+}
+
 int openconnect_set_reported_os(struct openconnect_info *vpninfo,
 				const char *os)
 {
@@ -156,7 +227,10 @@ static void free_optlist(struct oc_vpn_option *opt)
 void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 {
 	openconnect_close_https(vpninfo, 1);
-	dtls_shutdown(vpninfo);
+	if (vpninfo->proto.udp_shutdown)
+		vpninfo->proto.udp_shutdown(vpninfo);
+	if (vpninfo->tncc_fd != -1)
+		closesocket(vpninfo->tncc_fd);
 	if (vpninfo->cmd_fd_write != -1) {
 		closesocket(vpninfo->cmd_fd);
 		closesocket(vpninfo->cmd_fd_write);
@@ -183,7 +257,7 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 	free_optlist(vpninfo->cookies);
 	free_optlist(vpninfo->cstp_options);
 	free_optlist(vpninfo->dtls_options);
-	cstp_free_splits(vpninfo);
+	free_split_routes(vpninfo);
 	free(vpninfo->hostname);
 	free(vpninfo->unique_hostname);
 	free(vpninfo->urlpath);
@@ -258,7 +332,6 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 	if (vpninfo->stoken_ctx)
 		stoken_destroy(vpninfo->stoken_ctx);
 #endif
-#ifdef HAVE_LIBOATH
 	if (vpninfo->oath_secret) {
 #ifdef HAVE_LIBPSKC
 		if (vpninfo->pskc)
@@ -266,9 +339,7 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 		else
 #endif /* HAVE_LIBPSKC */
 		free(vpninfo->oath_secret);
-		oath_done();
 	}
-#endif /* HAVE_LIBOATH */
 #ifdef HAVE_LIBPCSCLITE
 	if (vpninfo->token_mode == OC_TOKEN_MODE_YUBIOATH) {
 		SCardDisconnect(vpninfo->pcsc_card, SCARD_LEAVE_CARD);
@@ -571,11 +642,7 @@ int openconnect_has_stoken_support(void)
 
 int openconnect_has_oath_support(void)
 {
-#ifdef HAVE_LIBOATH
 	return 2;
-#else
-	return 0;
-#endif
 }
 
 int openconnect_has_yubioath_support(void)
@@ -640,13 +707,11 @@ int openconnect_set_token_mode(struct openconnect_info *vpninfo,
 	case OC_TOKEN_MODE_STOKEN:
 		return set_libstoken_mode(vpninfo, token_str);
 #endif
-#ifdef HAVE_LIBOATH
 	case OC_TOKEN_MODE_TOTP:
 		return set_totp_mode(vpninfo, token_str);
 
 	case OC_TOKEN_MODE_HOTP:
 		return set_hotp_mode(vpninfo, token_str);
-#endif
 #ifdef HAVE_LIBPCSCLITE
 	case OC_TOKEN_MODE_YUBIOATH:
 		return set_yubikey_mode(vpninfo, token_str);
@@ -824,3 +889,97 @@ const char *openconnect_get_peer_cert_hash(struct openconnect_info *vpninfo)
 {
 	return vpninfo->peer_cert_hash;
 }
+
+int openconnect_set_compression_mode(struct openconnect_info *vpninfo,
+				     oc_compression_mode_t mode)
+{
+	switch(mode) {
+	case OC_COMPRESSION_MODE_NONE:
+		vpninfo->req_compr = 0;
+		return 0;
+	case OC_COMPRESSION_MODE_STATELESS:
+		vpninfo->req_compr = COMPR_STATELESS;
+		return 0;
+	case OC_COMPRESSION_MODE_ALL:
+		vpninfo->req_compr = COMPR_ALL;
+		return 0;
+	default:
+		return -EINVAL;
+	}
+}
+
+void nuke_opt_values(struct oc_form_opt *opt)
+{
+	for (; opt; opt = opt->next) {
+		if (opt->type == OC_FORM_OPT_TEXT ||
+		    opt->type == OC_FORM_OPT_PASSWORD) {
+			free(opt->_value);
+			opt->_value = NULL;
+		}
+	}
+}
+
+int process_auth_form(struct openconnect_info *vpninfo, struct oc_auth_form *form)
+{
+	int ret;
+	struct oc_form_opt_select *grp = form->authgroup_opt;
+	struct oc_choice *auth_choice;
+	struct oc_form_opt *opt;
+
+	if (!vpninfo->process_auth_form) {
+		vpn_progress(vpninfo, PRG_ERR, _("No form handler; cannot authenticate.\n"));
+		return OC_FORM_RESULT_ERR;
+	}
+
+retry:
+	auth_choice = NULL;
+	if (grp && grp->nr_choices && !vpninfo->xmlpost) {
+		if (vpninfo->authgroup) {
+			/* For non-XML-POST, the server doesn't tell us which group is selected */
+			int i;
+			for (i = 0; i < grp->nr_choices; i++)
+				if (!strcmp(grp->choices[i]->name, vpninfo->authgroup))
+					form->authgroup_selection = i;
+		}
+		auth_choice = grp->choices[form->authgroup_selection];
+	}
+
+	for (opt = form->opts; opt; opt = opt->next) {
+		int second_auth = opt->flags & OC_FORM_OPT_SECOND_AUTH;
+		opt->flags &= ~OC_FORM_OPT_IGNORE;
+
+		if (!auth_choice ||
+		    (opt->type != OC_FORM_OPT_TEXT && opt->type != OC_FORM_OPT_PASSWORD))
+			continue;
+
+		if (auth_choice->noaaa ||
+		    (!auth_choice->second_auth && second_auth))
+			opt->flags |= OC_FORM_OPT_IGNORE;
+		else if (!strcmp(opt->name, "secondary_username") && second_auth) {
+			if (auth_choice->secondary_username) {
+				free(opt->_value);
+				opt->_value = strdup(auth_choice->secondary_username);
+			}
+			if (!auth_choice->secondary_username_editable)
+				opt->flags |= OC_FORM_OPT_IGNORE;
+		}
+	}
+
+	ret = vpninfo->process_auth_form(vpninfo->cbdata, form);
+
+	if (ret == OC_FORM_RESULT_NEWGROUP &&
+	    form->authgroup_opt &&
+	    form->authgroup_opt->form._value) {
+		free(vpninfo->authgroup);
+		vpninfo->authgroup = strdup(form->authgroup_opt->form._value);
+
+		if (!vpninfo->xmlpost)
+			goto retry;
+	}
+
+	if (ret == OC_FORM_RESULT_CANCELLED || ret < 0)
+		nuke_opt_values(form->opts);
+
+	return ret;
+}
+
