@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2014 Intel Corporation.
+ * Copyright © 2008-2015 Intel Corporation.
  * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
  * Copyright © 2013 John Morrissey <jwm@horde.net>
  *
@@ -148,6 +148,7 @@ enum {
 	OPT_AUTHGROUP,
 	OPT_BASEMTU,
 	OPT_CAFILE,
+	OPT_COMPRESSION,
 	OPT_CONFIGFILE,
 	OPT_COOKIEONLY,
 	OPT_COOKIE_ON_STDIN,
@@ -158,6 +159,7 @@ enum {
 	OPT_DUMP_HTTP,
 	OPT_FORCE_DPD,
 	OPT_GNUTLS_DEBUG,
+	OPT_JUNIPER,
 	OPT_KEY_PASSWORD_FROM_FSID,
 	OPT_LIBPROXY,
 	OPT_NO_CERT_CHECK,
@@ -181,6 +183,7 @@ enum {
 	OPT_TIMESTAMP,
 	OPT_PFS,
 	OPT_PROXY_AUTH,
+	OPT_HTTP_AUTH,
 };
 
 #ifdef __sun__
@@ -207,11 +210,14 @@ static const struct option long_options[] = {
 	OPTION("certificate", 1, 'c'),
 	OPTION("sslkey", 1, 'k'),
 	OPTION("cookie", 1, 'C'),
+	OPTION("compression", 1, OPT_COMPRESSION),
 	OPTION("deflate", 0, 'd'),
+	OPTION("juniper", 0, OPT_JUNIPER),
 	OPTION("no-deflate", 0, 'D'),
 	OPTION("cert-expire-warning", 1, 'e'),
 	OPTION("usergroup", 1, 'g'),
 	OPTION("help", 0, 'h'),
+	OPTION("http-auth", 1, OPT_HTTP_AUTH),
 	OPTION("interface", 1, 'i'),
 	OPTION("mtu", 1, 'm'),
 	OPTION("base-mtu", 1, OPT_BASEMTU),
@@ -768,9 +774,6 @@ static void usage(void)
 #ifndef HAVE_LIBSTOKEN
 	printf("                                  %s\n", _("(NOTE: libstoken (RSA SecurID) disabled in this build)"));
 #endif
-#ifndef HAVE_LIBOATH
-	printf("                                  %s\n", _("(NOTE: liboath (TOTP,HOTP) disabled in this build)"));
-#endif
 #ifndef HAVE_LIBPCSCLITE
 	printf("                                  %s\n", _("(NOTE: Yubikey OATH disabled in this build)"));
 #endif
@@ -931,13 +934,13 @@ int main(int argc, char **argv)
 {
 	struct openconnect_info *vpninfo;
 	char *urlpath = NULL;
+	const char *compr = "";
 	char *proxy = getenv("https_proxy");
 	char *vpnc_script = NULL, *ifname = NULL;
 	const struct oc_ip_info *ip_info;
 	int autoproxy = 0;
 	int opt;
 	char *pidfile = NULL;
-	int use_dtls = 1;
 	FILE *fp = NULL;
 	char *config_arg;
 	char *config_filename;
@@ -1031,6 +1034,11 @@ int main(int argc, char **argv)
 			}
 			break;
 		}
+		case OPT_JUNIPER:
+			fprintf(stderr, "WARNING: Juniper Network Connect support is experimental.\n");
+			fprintf(stderr, "It will probably be superseded by Junos Pulse support.\n");
+			openconnect_set_protocol(vpninfo, "nc");
+			break;
 		case OPT_CSD_USER: {
 			char *strend;
 			vpninfo->uid_csd = strtol(config_arg, &strend, 0);
@@ -1067,6 +1075,20 @@ int main(int argc, char **argv)
 			config_line_num = 1;
 			/* The next option will come from the file... */
 			break;
+		case OPT_COMPRESSION:
+			if (!strcmp(config_arg, "none") ||
+			    !strcmp(config_arg, "off"))
+				openconnect_set_compression_mode(vpninfo, OC_COMPRESSION_MODE_NONE);
+			else if (!strcmp(config_arg, "all"))
+				openconnect_set_compression_mode(vpninfo, OC_COMPRESSION_MODE_ALL);
+			else if (!strcmp(config_arg, "stateless"))
+				openconnect_set_compression_mode(vpninfo, OC_COMPRESSION_MODE_STATELESS);
+			else {
+				fprintf(stderr, _("Invalid compression mode '%s'\n"),
+					config_arg);
+				exit(1);
+			}
+			break;
 		case OPT_CAFILE:
 			openconnect_set_cafile(vpninfo, dup_config_arg());
 			break;
@@ -1081,7 +1103,7 @@ int main(int argc, char **argv)
 			openconnect_set_system_trust(vpninfo, 0);
 			break;
 		case OPT_NO_DTLS:
-			use_dtls = 0;
+			vpninfo->dtls_state = DTLS_DISABLED;
 			break;
 		case OPT_COOKIEONLY:
 			cookieonly = 1;
@@ -1170,7 +1192,10 @@ int main(int argc, char **argv)
 			autoproxy = 0;
 			break;
 		case OPT_PROXY_AUTH:
-			openconnect_set_proxy_auth(vpninfo, dup_config_arg());
+			openconnect_set_proxy_auth(vpninfo, config_arg);
+			break;
+		case OPT_HTTP_AUTH:
+			openconnect_set_http_auth(vpninfo, config_arg);
 			break;
 		case OPT_NO_PROXY:
 			autoproxy = 0;
@@ -1419,18 +1444,34 @@ int main(int argc, char **argv)
 	}
 #endif
 
-	if (use_dtls && openconnect_setup_dtls(vpninfo, 60))
+	if (vpninfo->dtls_state != DTLS_DISABLED &&
+	    openconnect_setup_dtls(vpninfo, 60))
 		fprintf(stderr, _("Set up DTLS failed; using SSL instead\n"));
 
 	openconnect_get_ip_info(vpninfo, &ip_info, NULL, NULL);
+
+	if (vpninfo->dtls_state != DTLS_CONNECTED) {
+		if (vpninfo->cstp_compr == COMPR_DEFLATE)
+			compr = " + deflate";
+		else if (vpninfo->cstp_compr == COMPR_LZS)
+			compr = " + lzs";
+		else if (vpninfo->cstp_compr == COMPR_LZ4)
+			compr = " + lz4";
+	} else {
+		if (vpninfo->dtls_compr == COMPR_DEFLATE)
+			compr = " + deflate";
+		else if (vpninfo->dtls_compr == COMPR_LZS)
+			compr = " + lzs";
+		else if (vpninfo->dtls_compr == COMPR_LZ4)
+			compr = " + lz4";
+	}
 	vpn_progress(vpninfo, PRG_INFO,
-		     _("Connected %s as %s%s%s, using %s\n"), openconnect_get_ifname(vpninfo),
+		     _("Connected %s as %s%s%s, using %s%s\n"), openconnect_get_ifname(vpninfo),
 		     ip_info->addr?:"",
 		     (ip_info->netmask6 && ip_info->addr) ? " + " : "",
 		     ip_info->netmask6 ? : "",
-		     (vpninfo->dtls_state != DTLS_CONNECTED) ?
-		     (vpninfo->cstp_compr == COMPR_DEFLATE) ? "SSL + deflate" : "SSL"
-		     : "DTLS");
+		     (vpninfo->dtls_state != DTLS_CONNECTED) ? "SSL"
+		     : "DTLS", compr);
 
 	if (!vpninfo->vpnc_script) {
 		vpn_progress(vpninfo, PRG_INFO,
@@ -1471,6 +1512,8 @@ int main(int argc, char **argv)
 	}
 #endif
 
+	openconnect_set_loglevel(vpninfo, verbose);
+
 	while (1) {
 		ret = openconnect_mainloop(vpninfo, reconnect_timeout, RECONNECT_INTERVAL_MIN);
 		if (ret)
@@ -1823,6 +1866,11 @@ static int process_auth_form_cb(void *_vpninfo,
 			if (!opt->_value)
 				goto err;
 			empty = 0;
+		} else if (opt->type == OC_FORM_OPT_TOKEN) {
+			/* Nothing to do here, but if the tokencode is being
+			 * automatically generated then don't treat it as an
+			 * empty form for the purpose of loop avoidance. */
+			empty = 0;
 		}
 	}
 
@@ -1895,6 +1943,7 @@ static void init_token(struct openconnect_info *vpninfo,
 		switch(token_str[0]) {
 		case '@':
 			token_str++;
+			/* fall through... */
 		case '/':
 			if (read_file_into_string(vpninfo, token_str,
 						  &file_token) < 0)
