@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2014 Intel Corporation.
+ * Copyright © 2008-2015 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -25,16 +25,7 @@
 
 #include "openconnect-internal.h"
 
-void queue_packet(struct pkt **q, struct pkt *new)
-{
-	while (*q)
-		q = &(*q)->next;
-
-	new->next = NULL;
-	*q = new;
-}
-
-int queue_new_packet(struct pkt **q, void *buf, int len)
+int queue_new_packet(struct pkt_q *q, void *buf, int len)
 {
 	struct pkt *new = malloc(sizeof(struct pkt) + len);
 	if (!new)
@@ -51,6 +42,7 @@ int queue_new_packet(struct pkt **q, void *buf, int len)
    tun*.c files for specific platforms */
 int tun_mainloop(struct openconnect_info *vpninfo, int *timeout)
 {
+	struct pkt *this;
 	int work_done = 0;
 
 	if (read_fd_monitored(vpninfo, tun)) {
@@ -59,7 +51,7 @@ int tun_mainloop(struct openconnect_info *vpninfo, int *timeout)
 			int len = vpninfo->ip_info.mtu;
 
 			if (!out_pkt) {
-				out_pkt = malloc(sizeof(struct pkt) + len);
+				out_pkt = malloc(sizeof(struct pkt) + len + vpninfo->pkt_trailer);
 				if (!out_pkt) {
 					vpn_progress(vpninfo, PRG_ERR, _("Allocation failed\n"));
 					break;
@@ -72,33 +64,31 @@ int tun_mainloop(struct openconnect_info *vpninfo, int *timeout)
 
 			vpninfo->stats.tx_pkts++;
 			vpninfo->stats.tx_bytes += out_pkt->len;
-
-			queue_packet(&vpninfo->outgoing_queue, out_pkt);
-			out_pkt = NULL;
-
 			work_done = 1;
-			vpninfo->outgoing_qlen++;
-			if (vpninfo->outgoing_qlen == vpninfo->max_qlen) {
+
+			if (queue_packet(&vpninfo->outgoing_queue, out_pkt) ==
+			    vpninfo->max_qlen) {
+				out_pkt = NULL;
 				unmonitor_read_fd(vpninfo, tun);
 				break;
 			}
+			out_pkt = NULL;
 		}
 		vpninfo->tun_pkt = out_pkt;
-	} else if (vpninfo->outgoing_qlen < vpninfo->max_qlen) {
+	} else if (vpninfo->outgoing_queue.count < vpninfo->max_qlen) {
 		monitor_read_fd(vpninfo, tun);
 	}
 
-	while (vpninfo->incoming_queue) {
-		struct pkt *this = vpninfo->incoming_queue;
+	while ((this = dequeue_packet(&vpninfo->incoming_queue))) {
 
-		if (os_write_tun(vpninfo, this))
+		if (os_write_tun(vpninfo, this)) {
+			requeue_packet(&vpninfo->incoming_queue, this);
 			break;
+		}
 
 		vpninfo->stats.rx_pkts++;
 		vpninfo->stats.rx_bytes += this->len;
 
-		vpninfo->incoming_queue = this->next;
-
 		free(this);
 	}
 	/* Work is not done if we just got rid of packets off the queue */
@@ -138,16 +128,14 @@ int openconnect_mainloop(struct openconnect_info *vpninfo,
 		fd_set rfds, wfds, efds;
 #endif
 
-#ifdef HAVE_DTLS
-		if (vpninfo->dtls_state != DTLS_DISABLED) {
-			ret = dtls_mainloop(vpninfo, &timeout);
+		if (vpninfo->dtls_state > DTLS_DISABLED) {
+			ret = vpninfo->proto.udp_mainloop(vpninfo, &timeout);
 			if (vpninfo->quit_reason)
 				break;
 			did_work += ret;
 		}
-#endif
 
-		ret = cstp_mainloop(vpninfo, &timeout);
+		ret = vpninfo->proto.tcp_mainloop(vpninfo, &timeout);
 		if (vpninfo->quit_reason)
 			break;
 		did_work += ret;
@@ -173,7 +161,7 @@ int openconnect_mainloop(struct openconnect_info *vpninfo,
 			   openconnect_mainloop() again */
 			openconnect_close_https(vpninfo, 0);
 			if (vpninfo->dtls_state != DTLS_DISABLED) {
-				dtls_close(vpninfo);
+				vpninfo->proto.udp_close(vpninfo);
 				vpninfo->dtls_state = DTLS_SLEEPING;
 				vpninfo->new_dtls_started = 0;
 			}
@@ -224,8 +212,8 @@ int openconnect_mainloop(struct openconnect_info *vpninfo,
 #endif
 	}
 
-	if (vpninfo->quit_reason)
-		cstp_bye(vpninfo, vpninfo->quit_reason);
+	if (vpninfo->quit_reason && vpninfo->proto.vpn_close_session)
+		vpninfo->proto.vpn_close_session(vpninfo, vpninfo->quit_reason);
 
 	os_shutdown_tun(vpninfo);
 	return ret < 0 ? ret : -EIO;
