@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2014 Intel Corporation.
+ * Copyright © 2008-2015 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -37,13 +37,14 @@
 
 #include "openconnect-internal.h"
 
+#define auth_is_proxy(v, a) ((unsigned long)(a) >= ((unsigned long)(v)->proxy_auth))
 
 #define NTLM_SSO_REQ		2	/* SSO type1 packet sent */
 #define NTLM_MANUAL		3	/* SSO challenge/response sent or skipped; manual next */
 #define NTLM_MANUAL_REQ		4	/* manual type1 packet sent */
 
 #ifdef _WIN32
-static int ntlm_sspi(struct openconnect_info *vpninfo, struct oc_text_buf *buf, const char *challenge)
+static int ntlm_sspi(struct openconnect_info *vpninfo, struct http_auth_state *auth_state, struct oc_text_buf *buf, const char *challenge)
 {
         SECURITY_STATUS status;
         SecBufferDesc input_desc, output_desc;
@@ -72,13 +73,13 @@ static int ntlm_sspi(struct openconnect_info *vpninfo, struct oc_text_buf *buf,
         out_token.cbBuffer = 0;
         out_token.pvBuffer = NULL;
 
-	status = InitializeSecurityContextW(&vpninfo->ntlm_sspi_cred,
-					    challenge ? &vpninfo->ntlm_sspi_ctx : NULL,
+	status = InitializeSecurityContextW(&auth_state->ntlm_sspi_cred,
+					    challenge ? &auth_state->ntlm_sspi_ctx : NULL,
 					    (SEC_WCHAR *)L"",
 					    ISC_REQ_ALLOCATE_MEMORY | ISC_REQ_CONFIDENTIALITY | ISC_REQ_REPLAY_DETECT | ISC_REQ_CONNECTION,
 					    0, SECURITY_NETWORK_DREP,
 					    challenge ? &input_desc : NULL,
-					    0, &vpninfo->ntlm_sspi_ctx,
+					    0, &auth_state->ntlm_sspi_ctx,
 					    &output_desc, &ret_flags, NULL);
 
 	if (status != SEC_E_OK && status != SEC_I_CONTINUE_NEEDED) {
@@ -87,7 +88,8 @@ static int ntlm_sspi(struct openconnect_info *vpninfo, struct oc_text_buf *buf,
 		return -EIO;
 	}
 
-	buf_append(buf, "Proxy-Authorization: NTLM ");
+	buf_append(buf, "%sAuthorization: NTLM ",
+		   auth_is_proxy(vpninfo, auth_state) ? "Proxy-" : "");
 	buf_append_base64(buf, out_token.pvBuffer, out_token.cbBuffer);
 	buf_append(buf, "\r\n");
 
@@ -96,7 +98,8 @@ static int ntlm_sspi(struct openconnect_info *vpninfo, struct oc_text_buf *buf,
 	return 0;
 }
 
-static int ntlm_helper_spawn(struct openconnect_info *vpninfo, struct oc_text_buf *buf)
+static int ntlm_helper_spawn(struct openconnect_info *vpninfo, struct http_auth_state *auth_state,
+			     struct oc_text_buf *buf)
 {
         SECURITY_STATUS status;
 	int ret;
@@ -104,36 +107,40 @@ static int ntlm_helper_spawn(struct openconnect_info *vpninfo, struct oc_text_bu
 	status = AcquireCredentialsHandleW(NULL, (SEC_WCHAR *)L"NTLM",
 					   SECPKG_CRED_OUTBOUND, NULL, NULL,
 					   NULL, NULL,
-					   &vpninfo->ntlm_sspi_cred, NULL);
+					   &auth_state->ntlm_sspi_cred, NULL);
 	if (status != SEC_E_OK) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("AcquireCredentialsHandle() failed: %lx\n"), status);
 		return -EIO;
 	}
 
-	ret = ntlm_sspi(vpninfo, buf, NULL);
+	ret = ntlm_sspi(vpninfo, auth_state, buf, NULL);
 	if (ret)
-		FreeCredentialsHandle(&vpninfo->ntlm_sspi_cred);
+		FreeCredentialsHandle(&auth_state->ntlm_sspi_cred);
 
 	return ret;
 }
 
-static int ntlm_helper_challenge(struct openconnect_info *vpninfo, struct oc_text_buf *buf)
+static int ntlm_helper_challenge(struct openconnect_info *vpninfo,
+				 struct http_auth_state *auth_state,
+				 struct oc_text_buf *buf)
 {
-	return ntlm_sspi(vpninfo, buf, vpninfo->auth[AUTH_TYPE_NTLM].challenge);
+	return ntlm_sspi(vpninfo, auth_state, buf, auth_state->challenge);
 }
 
-void cleanup_ntlm_auth(struct openconnect_info *vpninfo)
+void cleanup_ntlm_auth(struct openconnect_info *vpninfo,
+		       struct http_auth_state *auth_state)
 {
-	if (vpninfo->auth[AUTH_TYPE_NTLM].state == NTLM_SSO_REQ) {
-		FreeCredentialsHandle(&vpninfo->ntlm_sspi_cred);
-		DeleteSecurityContext(&vpninfo->ntlm_sspi_ctx);
+	if (auth_state->state == NTLM_SSO_REQ) {
+		FreeCredentialsHandle(&auth_state->ntlm_sspi_cred);
+		DeleteSecurityContext(&auth_state->ntlm_sspi_ctx);
 	}
 }
 
 #else /* !_WIN32 */
 
-static int ntlm_helper_spawn(struct openconnect_info *vpninfo, struct oc_text_buf *buf)
+static int ntlm_helper_spawn(struct openconnect_info *vpninfo, struct http_auth_state *auth_state,
+			     struct oc_text_buf *buf)
 {
 	char *username;
 	int pipefd[2];
@@ -216,28 +223,32 @@ static int ntlm_helper_spawn(struct openconnect_info *vpninfo, struct oc_text_bu
 		return -EIO;
 	}
 	helperbuf[len - 1] = 0;
-	buf_append(buf, "Proxy-Authorization: NTLM %s\r\n", helperbuf + 3);
-	vpninfo->ntlm_helper_fd = pipefd[1];
+	buf_append(buf, "%sAuthorization: NTLM %s\r\n",
+		   auth_is_proxy(vpninfo, auth_state) ? "Proxy-" : "",
+		   helperbuf + 3);
+	auth_state->ntlm_helper_fd = pipefd[1];
 	return 0;
 }
 
-static int ntlm_helper_challenge(struct openconnect_info *vpninfo, struct oc_text_buf *buf)
+static int ntlm_helper_challenge(struct openconnect_info *vpninfo,
+				 struct http_auth_state *auth_state,
+				 struct oc_text_buf *buf)
 {
 	char helperbuf[4096];
 	int len;
 
-	if (!vpninfo->auth[AUTH_TYPE_NTLM].challenge ||
-	    write(vpninfo->ntlm_helper_fd, "TT ", 3) != 3 ||
-	    write(vpninfo->ntlm_helper_fd, vpninfo->auth[AUTH_TYPE_NTLM].challenge,
-		  strlen(vpninfo->auth[AUTH_TYPE_NTLM].challenge)) != strlen(vpninfo->auth[AUTH_TYPE_NTLM].challenge) ||
-	    write(vpninfo->ntlm_helper_fd, "\n", 1) != 1) {
+	if (!auth_state->challenge ||
+	    write(auth_state->ntlm_helper_fd, "TT ", 3) != 3 ||
+	    write(auth_state->ntlm_helper_fd, auth_state->challenge,
+		  strlen(auth_state->challenge)) != strlen(auth_state->challenge) ||
+	    write(auth_state->ntlm_helper_fd, "\n", 1) != 1) {
 	err:
 		vpn_progress(vpninfo, PRG_ERR, _("Error communicating with ntlm_auth helper\n"));
-		close(vpninfo->ntlm_helper_fd);
-		vpninfo->ntlm_helper_fd = -1;
+		close(auth_state->ntlm_helper_fd);
+		auth_state->ntlm_helper_fd = -1;
 		return -EAGAIN;
 	}
-	len = read(vpninfo->ntlm_helper_fd, helperbuf, sizeof(helperbuf));
+	len = read(auth_state->ntlm_helper_fd, helperbuf, sizeof(helperbuf));
 	/* Accept both 'KK' and 'AF'. It should be the latter but see
 	   https://bugzilla.samba.org/show_bug.cgi?id=10691 */
 	if (len < 4 || (!(helperbuf[0] == 'K' && helperbuf[1] == 'K') &&
@@ -246,18 +257,26 @@ static int ntlm_helper_challenge(struct openconnect_info *vpninfo, struct oc_tex
 		goto err;
 	}
 	helperbuf[len - 1] = 0;
-	buf_append(buf, "Proxy-Authorization: NTLM %s\r\n", helperbuf + 3);
+	buf_append(buf, "%sAuthorization: NTLM %s\r\n",
+		   auth_is_proxy(vpninfo, auth_state) ? "Proxy-" : "", helperbuf + 3);
 
-	vpn_progress(vpninfo, PRG_INFO, _("Attempting HTTP NTLM authentication to proxy (single-sign-on)\n"));
+	if (auth_is_proxy(vpninfo, auth_state))
+		vpn_progress(vpninfo, PRG_INFO,
+			     _("Attempting HTTP NTLM authentication to proxy (single-sign-on)\n"));
+	else
+		vpn_progress(vpninfo, PRG_INFO,
+			     _("Attempting HTTP NTLM authentication to server '%s' (single-sign-on)\n"),
+			     vpninfo->hostname);
 	return 0;
 
 }
 
-void cleanup_ntlm_auth(struct openconnect_info *vpninfo)
+void cleanup_ntlm_auth(struct openconnect_info *vpninfo,
+		       struct http_auth_state *auth_state)
 {
-	if (vpninfo->auth[AUTH_TYPE_NTLM].state == NTLM_SSO_REQ) {
-		close(vpninfo->ntlm_helper_fd);
-		vpninfo->ntlm_helper_fd = -1;
+	if (auth_state->state == NTLM_SSO_REQ) {
+		close(auth_state->ntlm_helper_fd);
+		auth_state->ntlm_helper_fd = -1;
 	}
 }
 #endif /* !_WIN32 */
@@ -299,10 +318,7 @@ static int md4sum (struct oc_text_buf *buf, unsigned char digest[16])
 	M = (void *)buf->data;
 	memset (M + nbytes, 0, pbytes + 8);
 	M[nbytes] = 0x80;
-	M[nbytes + pbytes] = nbits & 0xFF;
-	M[nbytes + pbytes + 1] = (nbits >> 8) & 0xFF;
-	M[nbytes + pbytes + 2] = (nbits >> 16) & 0xFF;
-	M[nbytes + pbytes + 3] = (nbits >> 24) & 0xFF;
+	store_le32(&M[nbytes + pbytes], nbits);
 
 	A = 0x67452301;
 	B = 0xEFCDAB89;
@@ -310,12 +326,8 @@ static int md4sum (struct oc_text_buf *buf, unsigned char digest[16])
 	D = 0x10325476;
 
 	for (i = 0; i < nbytes + pbytes + 8; i += 64) {
-		for (j = 0; j < 16; j++) {
-			X[j] =  (M[i + j * 4]) |
-				(M[i + j * 4 + 1] << 8) |
-				(M[i + j * 4 + 2] << 16) |
-				(M[i + j * 4 + 3] << 24);
-		}
+		for (j = 0; j < 16; j++)
+			X[j] =  load_le32(&M[i + j * 4]);
 
 		AA = A;
 		BB = B;
@@ -379,22 +391,10 @@ static int md4sum (struct oc_text_buf *buf, unsigned char digest[16])
 		D += DD;
 	}
 
-	digest[0]  =  A        & 0xFF;
-	digest[1]  = (A >>  8) & 0xFF;
-	digest[2]  = (A >> 16) & 0xFF;
-	digest[3]  = (A >> 24) & 0xFF;
-	digest[4]  =  B        & 0xFF;
-	digest[5]  = (B >>  8) & 0xFF;
-	digest[6]  = (B >> 16) & 0xFF;
-	digest[7]  = (B >> 24) & 0xFF;
-	digest[8]  =  C        & 0xFF;
-	digest[9]  = (C >>  8) & 0xFF;
-	digest[10] = (C >> 16) & 0xFF;
-	digest[11] = (C >> 24) & 0xFF;
-	digest[12] =  D        & 0xFF;
-	digest[13] = (D >>  8) & 0xFF;
-	digest[14] = (D >> 16) & 0xFF;
-	digest[15] = (D >> 24) & 0xFF;
+	store_le32(digest,      A);
+	store_le32(digest + 4,  B);
+	store_le32(digest + 8,  C);
+	store_le32(digest + 12, D);
 
 	return 0;
 }
@@ -551,14 +551,8 @@ static void des (uint32_t ks[16][2], unsigned char block[8])
 	uint32_t left, right, work;
 
 	/* Read input block and place in left/right in big-endian order */
-	left = ((uint32_t) block[0] << 24)
-	 | ((uint32_t) block[1] << 16)
-	 | ((uint32_t) block[2] << 8)
-	 | (uint32_t) block[3];
-	right = ((uint32_t) block[4] << 24)
-	 | ((uint32_t) block[5] << 16)
-	 | ((uint32_t) block[6] << 8)
-	 | (uint32_t) block[7];
+	left = load_be32(block);
+	right = load_be32(block + 4);
 
 	/* Hoey's clever initial permutation algorithm, from Outerbridge
 	 * (see Schneier p 478)
@@ -626,14 +620,8 @@ static void des (uint32_t ks[16][2], unsigned char block[8])
 	right ^= work << 4;
 
 	/* Put the block back into the user's buffer with final swap */
-	block[0] = right >> 24;
-	block[1] = right >> 16;
-	block[2] = right >> 8;
-	block[3] = right;
-	block[4] = left >> 24;
-	block[5] = left >> 16;
-	block[6] = left >> 8;
-	block[7] = left;
+	store_be32(block, right);
+	store_be32(block + 4, left);
 }
 
 /* Key schedule-related tables from FIPS-46 */
@@ -785,23 +773,24 @@ static int ntlm_nt_hash (const char *pass, char hash[21])
 	   to the next multiple of 64. */
 	ret = buf_ensure_space(utf16pass, ((strlen(pass) * 2) + 1 + 8 + 63) & ~63);
 	if (ret)
-		return ret;
+		goto out;
 
 	ret = buf_append_utf16le(utf16pass, pass);
 	if (ret < 0)
-		return ret;
+		goto wipe;
 
 	ret = buf_error(utf16pass);
 	if (ret)
-		return ret;
+		goto wipe;
 
 	ret = md4sum(utf16pass, (unsigned char *) hash);
 	if (ret)
-		return ret;
-
-	memset (hash + 16, 0, 5);
+		goto wipe;
 
+	memset(hash + 16, 0, 5);
+ wipe:
 	memset(utf16pass->data, 0, utf16pass->pos);
+ out:
 	buf_free(utf16pass);
 	return 0;
 }
@@ -826,32 +815,6 @@ static void ntlm_calc_response (const unsigned char key[21],
 	des (ks, results + 16);
 }
 
-static inline uint32_t load_le32(void *_p)
-{
-	unsigned char *p = _p;
-	return (p[3] << 24) | (p[2] << 16) | p[1] << 8 | p[0];
-}
-static inline uint16_t load_le16(void *_p)
-{
-	unsigned char *p = _p;
-	return p[1] << 8 | p[0];
-}
-
-static inline void store_le32(void *_p, uint32_t v)
-{
-	unsigned char *p = _p;
-	p[0] = v;
-	p[1] = v >> 8;
-	p[2] = v >> 16;
-	p[3] = v >> 24;
-}
-static inline void store_le16(void *_p, uint16_t v)
-{
-	unsigned char *p = _p;
-	p[0] = v;
-	p[1] = v >> 8;
-}
-
 #define NTLM_CHALLENGE_DOMAIN_OFFSET		12
 #define NTLM_CHALLENGE_FLAGS_OFFSET		20
 #define NTLM_CHALLENGE_NONCE_OFFSET		24
@@ -898,7 +861,10 @@ static void ntlm_set_string_binary(struct oc_text_buf *buf, int offset,
 	buf_append_bytes(buf, data, len);
 }
 
-static int ntlm_manual_challenge(struct openconnect_info *vpninfo, struct oc_text_buf *hdrbuf)
+static int ntlm_manual_challenge(struct openconnect_info *vpninfo,
+				 struct http_auth_state *auth_state,
+				 struct oc_text_buf *hdrbuf,
+				 const char *domuser, const char *pass)
 {
 	struct oc_text_buf *resp;
 	char *user;
@@ -907,14 +873,14 @@ static int ntlm_manual_challenge(struct openconnect_info *vpninfo, struct oc_tex
 	int token_len = -EINVAL;
 	int ntlmver;
 
-	if (!vpninfo->auth[AUTH_TYPE_NTLM].challenge)
+	if (!auth_state->challenge)
 		return -EINVAL;
 
-	if (ntlm_nt_hash (vpninfo->proxy_pass, (char *) hash))
+	if (ntlm_nt_hash (pass, (char *) hash))
 		return -EINVAL;
 
 	token = openconnect_base64_decode(&token_len,
-					  vpninfo->auth[AUTH_TYPE_NTLM].challenge);
+					  auth_state->challenge);
 	if (!token)
 		return token_len;
 
@@ -960,7 +926,7 @@ static int ntlm_manual_challenge(struct openconnect_info *vpninfo, struct oc_tex
 		ntlmver = 1;
 		memcpy (nonce, token + NTLM_CHALLENGE_NONCE_OFFSET, 8);
 		ntlm_calc_response (hash, nonce, nt_resp);
-		ntlm_lanmanager_hash (vpninfo->proxy_pass, (char *) hash);
+		ntlm_lanmanager_hash (pass, (char *) hash);
 		ntlm_calc_response (hash, nonce, lm_resp);
 	}
 
@@ -973,10 +939,10 @@ static int ntlm_manual_challenge(struct openconnect_info *vpninfo, struct oc_tex
 	/* Mask in the NTLM2SESSION flag */
 	resp->data[NTLM_RESPONSE_FLAGS_OFFSET + 2] = token[NTLM_CHALLENGE_FLAGS_OFFSET + 2] & 8;
 
-	user = strchr(vpninfo->proxy_user, '\\');
+	user = strchr(domuser, '\\');
 	if (user) {
 		*user = 0;
-		ntlm_set_string_utf8(resp, NTLM_RESPONSE_DOMAIN_OFFSET, vpninfo->proxy_user);
+		ntlm_set_string_utf8(resp, NTLM_RESPONSE_DOMAIN_OFFSET, domuser);
 		*user = '\\';
 		user++;
 	} else {
@@ -989,7 +955,7 @@ static int ntlm_manual_challenge(struct openconnect_info *vpninfo, struct oc_tex
 		}
 		ntlm_set_string_binary(resp, NTLM_RESPONSE_DOMAIN_OFFSET, token + offset, len);
 
-		user = vpninfo->proxy_user;
+		user = (char *)domuser;
 	}
 
 	ntlm_set_string_utf8(resp, NTLM_RESPONSE_USER_OFFSET, user);
@@ -1002,54 +968,70 @@ static int ntlm_manual_challenge(struct openconnect_info *vpninfo, struct oc_tex
 	if (buf_error(resp))
 		return buf_free(resp);
 
-	buf_append(hdrbuf, "Proxy-Authorization: NTLM ");
+	buf_append(hdrbuf, "%sAuthorization: NTLM ",
+		   auth_is_proxy(vpninfo, auth_state) ? "Proxy-" : "");
 	buf_append_base64(hdrbuf, resp->data, resp->pos);
 	buf_append(hdrbuf, "\r\n");
 
 	buf_free(resp);
-	vpn_progress(vpninfo, PRG_INFO,
-		     _("Attempting HTTP NTLMv%d authentication to proxy\n"),
-		     ntlmver);
+	if (auth_is_proxy(vpninfo, auth_state))
+		vpn_progress(vpninfo, PRG_INFO,
+			     _("Attempting HTTP NTLMv%d authentication to proxy\n"),
+			     ntlmver);
+	else
+		vpn_progress(vpninfo, PRG_INFO,
+			     _("Attempting HTTP NTLMv%d authentication to server '%s'\n"),
+			     ntlmver, vpninfo->hostname);
 	return 0;
 }
 
-int ntlm_authorization(struct openconnect_info *vpninfo, struct oc_text_buf *buf)
+int ntlm_authorization(struct openconnect_info *vpninfo, int proxy,
+		       struct http_auth_state *auth_state, struct oc_text_buf *buf)
 {
-	if (vpninfo->auth[AUTH_TYPE_NTLM].state == AUTH_AVAILABLE) {
-		vpninfo->auth[AUTH_TYPE_NTLM].state = NTLM_MANUAL;
+	const char *user, *pass;
+
+	if (proxy) {
+		user = vpninfo->proxy_user;
+		pass = vpninfo->proxy_pass;
+	} else {
+		user = pass = NULL;
+	}
+
+	if (auth_state->state == AUTH_AVAILABLE) {
+		auth_state->state = NTLM_MANUAL;
 		/* Don't attempt automatic NTLM auth if we were given a password */
-		if (!vpninfo->proxy_pass && !ntlm_helper_spawn(vpninfo, buf)) {
-			vpninfo->auth[AUTH_TYPE_NTLM].state = NTLM_SSO_REQ;
+		if (!pass && !ntlm_helper_spawn(vpninfo, auth_state, buf)) {
+			auth_state->state = NTLM_SSO_REQ;
 			return 0;
 		}
 	}
-	if (vpninfo->auth[AUTH_TYPE_NTLM].state == NTLM_SSO_REQ) {
+	if (auth_state->state == NTLM_SSO_REQ) {
 		int ret;
-		ret = ntlm_helper_challenge(vpninfo, buf);
+		ret = ntlm_helper_challenge(vpninfo, auth_state, buf);
 		/* Clean up after it. We're done here, whether it worked or not */
-		cleanup_ntlm_auth(vpninfo);
-		vpninfo->auth[AUTH_TYPE_NTLM].state = NTLM_MANUAL;
+		cleanup_ntlm_auth(vpninfo, auth_state);
+		auth_state->state = NTLM_MANUAL;
 		if (ret == -EAGAIN) {
 			/* Don't let it reset our state when it reconnects */
-			vpninfo->proxy_close_during_auth = 1;
+			if (proxy)
+				vpninfo->proxy_close_during_auth = 1;
 			return ret;
 		}
 		if (!ret)
 			return ret;
 	}
-	if (vpninfo->auth[AUTH_TYPE_NTLM].state == NTLM_MANUAL && vpninfo->proxy_user &&
-	    vpninfo->proxy_pass) {
-		buf_append(buf, "Proxy-Authorization: NTLM %s\r\n",
+	if (auth_state->state == NTLM_MANUAL && user && pass) {
+		buf_append(buf, "%sAuthorization: NTLM %s\r\n", proxy ? "Proxy-" : "",
 			   "TlRMTVNTUAABAAAABYIIAAAAAAAAAAAAAAAAAAAAAAAAAAAAMAAAAAAAAAAwAAAA");
-		vpninfo->auth[AUTH_TYPE_NTLM].state = NTLM_MANUAL_REQ;
+		auth_state->state = NTLM_MANUAL_REQ;
 		return 0;
 	}
-	if (vpninfo->auth[AUTH_TYPE_NTLM].state == NTLM_MANUAL_REQ &&
-	    !ntlm_manual_challenge(vpninfo, buf)) {
+	if (auth_state->state == NTLM_MANUAL_REQ && user && pass &&
+	    !ntlm_manual_challenge(vpninfo, auth_state, buf, user, pass)) {
 		/* Leave the state as it is. If we come back there'll be no
 		   challenge string and we'll fail then. */
 		return 0;
 	}
-	vpninfo->auth[AUTH_TYPE_NTLM].state = AUTH_FAILED;
+	auth_state->state = AUTH_FAILED;
 	return -EAGAIN;
 }
