@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2014 Intel Corporation.
+ * Copyright © 2008-2015 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -33,7 +33,8 @@
 #include <openssl/x509v3.h>
 #include <openssl/x509.h>
 #include <openssl/bio.h>
-
+#include <openssl/ui.h>
+#include <openssl/rsa.h>
 
 int openconnect_sha1(unsigned char *result, void *data, int len)
 {
@@ -215,6 +216,47 @@ static int openconnect_openssl_gets(struct openconnect_info *vpninfo, char *buf,
 	return i ?: ret;
 }
 
+int ssl_nonblock_read(struct openconnect_info *vpninfo, void *buf, int maxlen)
+{
+	int len, ret;
+
+	len = SSL_read(vpninfo->https_ssl, buf, maxlen);
+	if (len > 0)
+		return len;
+
+	ret = SSL_get_error(vpninfo->https_ssl, len);
+	if (ret == SSL_ERROR_SYSCALL || ret == SSL_ERROR_ZERO_RETURN) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("SSL read error %d (server probably closed connection); reconnecting.\n"),
+			     ret);
+		return -EIO;
+	}
+	return 0;
+}
+
+int ssl_nonblock_write(struct openconnect_info *vpninfo, void *buf, int buflen)
+{
+	int ret;
+
+	ret = SSL_write(vpninfo->https_ssl, buf, buflen);
+	if (ret > 0)
+		return ret;
+
+	ret = SSL_get_error(vpninfo->https_ssl, ret);
+	switch (ret) {
+	case SSL_ERROR_WANT_WRITE:
+		/* Waiting for the socket to become writable -- it's
+		   probably stalled, and/or the buffers are full */
+		monitor_write_fd(vpninfo, ssl);
+	case SSL_ERROR_WANT_READ:
+		return 0;
+
+	default:
+		vpn_progress(vpninfo, PRG_ERR, _("SSL_write failed: %d\n"), ret);
+		openconnect_report_ssl_errors(vpninfo);
+		return -1;
+	}
+}
 
 /* UI handling. All this just to handle the PIN callback from the TPM ENGINE,
    and turn it into a call to our ->process_auth_form function */
@@ -734,11 +776,11 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	FILE *f;
 	char buf[256];
 
-	if (!strncmp(vpninfo->sslkey, "pkcs11:", 7) ||
-	    !strncmp(vpninfo->cert, "pkcs11:", 7)) {
-		vpn_progress(vpninfo, PRG_ERR,
-			     _("This binary built without PKCS#11 support\n"));
-		return -EINVAL;
+	if (!strncmp(vpninfo->cert, "pkcs11:", 7)) {
+		int ret = load_pkcs11_certificate(vpninfo);
+		if (ret)
+			return ret;
+		goto got_cert;
 	}
 
 	vpn_progress(vpninfo, PRG_DEBUG,
@@ -792,6 +834,8 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		if (ret)
 			return ret;
 	}
+
+ got_cert:
 #ifdef ANDROID_KEYSTORE
 	if (!strncmp(vpninfo->sslkey, "keystore:", 9)) {
 		EVP_PKEY *key;
@@ -819,6 +863,8 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		return 0;
 	}
 #endif /* ANDROID_KEYSTORE */
+	if (!strncmp(vpninfo->sslkey, "pkcs11:", 7))
+		return load_pkcs11_key(vpninfo);
 
 	f = openconnect_fopen_utf8(vpninfo, vpninfo->sslkey, "rb");
 	if (!f) {
@@ -1229,6 +1275,12 @@ static int verify_peer(struct openconnect_info *vpninfo, SSL *https_ssl)
 	return ret;
 }
 
+/* Before OpenSSL 1.1 we could do this directly. And needed to. */
+#ifndef SSL_CTX_get_extra_chain_certs_only
+#define SSL_CTX_get_extra_chain_certs_only(ctx, st) \
+	do { *(st) = (ctx)->extra_certs; } while(0)
+#endif
+
 static void workaround_openssl_certchain_bug(struct openconnect_info *vpninfo,
 					     SSL *ssl)
 {
@@ -1240,12 +1292,14 @@ static void workaround_openssl_certchain_bug(struct openconnect_info *vpninfo,
 	X509 *cert2;
 	X509_STORE *store = SSL_CTX_get_cert_store(vpninfo->https_ctx);
 	X509_STORE_CTX ctx;
+	void *extra_certs;
 
 	if (!cert || !store)
 		return;
 
 	/* If we already have 'supporting' certs, don't add them again */
-	if (vpninfo->https_ctx->extra_certs)
+	SSL_CTX_get_extra_chain_certs_only(vpninfo->https_ctx, &extra_certs);
+	if (extra_certs)
 		return;
 
 	if (!X509_STORE_CTX_init(&ctx, store, NULL, NULL))
@@ -1602,9 +1656,10 @@ int openconnect_local_cert_md5(struct openconnect_info *vpninfo,
 
 #ifdef HAVE_LIBPCSCLITE
 int openconnect_hash_yubikey_password(struct openconnect_info *vpninfo,
-				      const char *password, const void *ident, int id_len)
+				      const char *password, int pwlen,
+				      const void *ident, int id_len)
 {
-	if (!PKCS5_PBKDF2_HMAC_SHA1(password, strlen(password), ident, id_len, 1000, 16,
+	if (!PKCS5_PBKDF2_HMAC_SHA1(password, pwlen, ident, id_len, 1000, 16,
 				    vpninfo->yubikey_pwhash))
 		return -EIO;
 
@@ -1622,3 +1677,36 @@ int openconnect_yubikey_chalresp(struct openconnect_info *vpninfo,
 	return 0;
 }
 #endif
+
+int hotp_hmac(struct openconnect_info *vpninfo, const void *challenge)
+{
+	unsigned char hash[64]; /* Enough for a SHA256 */
+	unsigned int hashlen = sizeof(hash);
+	const EVP_MD *alg;
+
+	switch(vpninfo->oath_hmac_alg) {
+	case OATH_ALG_HMAC_SHA1:
+		alg = EVP_sha1();
+		break;
+	case OATH_ALG_HMAC_SHA256:
+		alg = EVP_sha256();
+		break;
+	case OATH_ALG_HMAC_SHA512:
+		alg = EVP_sha512();
+		break;
+	default:
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Unsupported OATH HMAC algorithm\n"));
+		return -EINVAL;
+	}
+	if (!HMAC(alg, vpninfo->oath_secret, vpninfo->oath_secret_len,
+		  challenge, 8, hash, &hashlen)) {
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  _("Failed to calculate OATH HMAC\n"));
+		openconnect_report_ssl_errors(vpninfo);
+		return -EINVAL;
+	}
+
+	hashlen = hash[hashlen - 1] & 15;
+	return load_be32(&hash[hashlen]) & 0x7fffffff;
+}
