@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2014 Intel Corporation.
+ * Copyright © 2008-2015 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -166,6 +166,19 @@ static void setenv_cstp_opts(struct openconnect_info *vpninfo)
 	free(env_buf);
 }
 
+static unsigned char nybble(unsigned char n)
+{
+	if      (n >= '0' && n <= '9') return n - '0';
+	else if (n >= 'A' && n <= 'F') return n - ('A' - 10);
+	else if (n >= 'a' && n <= 'f') return n - ('a' - 10);
+	return 0;
+}
+
+unsigned char unhex(const char *data)
+{
+	return (nybble(data[0]) << 4) | nybble(data[1]);
+}
+
 static void set_banner(struct openconnect_info *vpninfo)
 {
 	char *banner, *legacy_banner, *q;
@@ -328,6 +341,30 @@ void prepare_script_env(struct openconnect_info *vpninfo)
 	setenv_cstp_opts(vpninfo);
 }
 
+void free_split_routes(struct openconnect_info *vpninfo)
+{
+	struct oc_split_include *inc;
+
+	for (inc = vpninfo->ip_info.split_includes; inc; ) {
+		struct oc_split_include *next = inc->next;
+		free(inc);
+		inc = next;
+	}
+	for (inc = vpninfo->ip_info.split_excludes; inc; ) {
+		struct oc_split_include *next = inc->next;
+		free(inc);
+		inc = next;
+	}
+	for (inc = vpninfo->ip_info.split_dns; inc; ) {
+		struct oc_split_include *next = inc->next;
+		free(inc);
+		inc = next;
+	}
+	vpninfo->ip_info.split_dns = vpninfo->ip_info.split_includes =
+		vpninfo->ip_info.split_excludes = NULL;
+}
+
+
 #ifdef _WIN32
 static wchar_t *create_script_env(struct openconnect_info *vpninfo)
 {
@@ -366,7 +403,8 @@ static wchar_t *create_script_env(struct openconnect_info *vpninfo)
 
 		/* See if we can find it in the existing environment */
 		for (i = 0; i < nr_envs; i++) {
-			if (!wcsncmp((wchar_t *)buf->data, oldenv[i], buf->pos / 2)) {
+			if (oldenv[i] &&
+			    !wcsncmp((wchar_t *)buf->data, oldenv[i], buf->pos / 2)) {
 				oldenv[i] = NULL;
 				break;
 			}
