@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2014 Intel Corporation.
+ * Copyright © 2008-2015 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -23,38 +23,39 @@
 #include "openconnect-internal.h"
 
 
-static int sspi_setup(struct openconnect_info *vpninfo, const char *service)
+static int sspi_setup(struct openconnect_info *vpninfo, struct http_auth_state *auth_state, const char *service, int proxy)
 {
 	SECURITY_STATUS status;
 	struct oc_text_buf *buf = buf_alloc();
 
 	buf_append_utf16le(buf, service);
 	buf_append_utf16le(buf, "/");
-	buf_append_utf16le(buf, vpninfo->proxy);
+	buf_append_utf16le(buf, proxy ? vpninfo->proxy : vpninfo->hostname);
 
 	if (buf_error(buf))
 		return buf_free(buf);
 
-	vpninfo->sspi_target_name = (wchar_t *)buf->data;
+	auth_state->sspi_target_name = (wchar_t *)buf->data;
 	buf->data = NULL;
 	buf_free(buf);
 
 	status = AcquireCredentialsHandleW(NULL, (SEC_WCHAR *)L"Negotiate",
 					   SECPKG_CRED_OUTBOUND, NULL, NULL,
-					   NULL, NULL, &vpninfo->sspi_cred,
+					   NULL, NULL, &auth_state->sspi_cred,
 					   NULL);
 	if (status != SEC_E_OK) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("AcquireCredentialsHandle() failed: %lx\n"), status);
-		free(vpninfo->sspi_target_name);
-		vpninfo->sspi_target_name = NULL;
+		free(auth_state->sspi_target_name);
+		auth_state->sspi_target_name = NULL;
 		return -EIO;
 	}
 
 	return 0;
 }
 
-int gssapi_authorization(struct openconnect_info *vpninfo, struct oc_text_buf *hdrbuf)
+int gssapi_authorization(struct openconnect_info *vpninfo, int proxy,
+			 struct http_auth_state *auth_state, struct oc_text_buf *hdrbuf)
 {
 	SECURITY_STATUS status;
 	SecBufferDesc input_desc, output_desc;
@@ -62,12 +63,12 @@ int gssapi_authorization(struct openconnect_info *vpninfo, struct oc_text_buf *h
 	ULONG ret_flags;
 	int first = 1;
 
-	if (vpninfo->auth[AUTH_TYPE_GSSAPI].state == AUTH_AVAILABLE && sspi_setup(vpninfo, "HTTP")) {
-		vpninfo->auth[AUTH_TYPE_GSSAPI].state = AUTH_FAILED;
+	if (auth_state->state == AUTH_AVAILABLE && sspi_setup(vpninfo, auth_state, "HTTP", proxy)) {
+		auth_state->state = AUTH_FAILED;
 		return -EIO;
 	}
 
-	if (vpninfo->auth[AUTH_TYPE_GSSAPI].challenge && *vpninfo->auth[AUTH_TYPE_GSSAPI].challenge) {
+	if (auth_state->challenge && *auth_state->challenge) {
 		int token_len = -EINVAL;
 
 		input_desc.cBuffers = 1;
@@ -76,21 +77,21 @@ int gssapi_authorization(struct openconnect_info *vpninfo, struct oc_text_buf *h
 
 		in_token.BufferType = SECBUFFER_TOKEN;
 		in_token.pvBuffer = openconnect_base64_decode(&token_len,
-							      vpninfo->auth[AUTH_TYPE_GSSAPI].challenge);
+							      auth_state->challenge);
 		if (!in_token.pvBuffer)
 			return token_len;
 		in_token.cbBuffer = token_len;
 
 		first = 0;
 
-	} else if (vpninfo->auth[AUTH_TYPE_GSSAPI].state > AUTH_AVAILABLE) {
+	} else if (auth_state->state > AUTH_AVAILABLE) {
 		/* This indicates failure. We were trying, but got an empty
 		   'Proxy-Authorization: Negotiate' header back from the server
 		   implying that we should start again... */
 		goto fail_gssapi;
 	}
 
-	vpninfo->auth[AUTH_TYPE_GSSAPI].state = AUTH_IN_PROGRESS;
+	auth_state->state = AUTH_IN_PROGRESS;
 
 	output_desc.cBuffers = 1;
 	output_desc.pBuffers = &out_token;
@@ -100,20 +101,20 @@ int gssapi_authorization(struct openconnect_info *vpninfo, struct oc_text_buf *h
 	out_token.cbBuffer = 0;
 	out_token.pvBuffer = NULL;
 
-	status = InitializeSecurityContextW(&vpninfo->sspi_cred,
-					    first ? NULL : &vpninfo->sspi_ctx,
-					    vpninfo->sspi_target_name,
+	status = InitializeSecurityContextW(&auth_state->sspi_cred,
+					    first ? NULL : &auth_state->sspi_ctx,
+					    auth_state->sspi_target_name,
 					    ISC_REQ_ALLOCATE_MEMORY | ISC_REQ_CONFIDENTIALITY | ISC_REQ_REPLAY_DETECT | ISC_REQ_CONNECTION,
 					    0, SECURITY_NETWORK_DREP,
 					    first ? NULL : &input_desc,
-					    0, &vpninfo->sspi_ctx,
+					    0, &auth_state->sspi_ctx,
 					    &output_desc, &ret_flags, NULL);
 	if (status != SEC_E_OK && status != SEC_I_CONTINUE_NEEDED) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("InitializeSecurityContext() failed: %lx\n"), status);
 	fail_gssapi:
-		cleanup_gssapi_auth(vpninfo);
-		vpninfo->auth[AUTH_TYPE_GSSAPI].state = AUTH_FAILED;
+		cleanup_gssapi_auth(vpninfo, auth_state);
+		auth_state->state = AUTH_FAILED;
 		/* -EAGAIN to first a reconnect if we had been trying. Else -EIO */
 		return first ? -EIO : -EAGAIN;
 	}
@@ -127,13 +128,15 @@ int gssapi_authorization(struct openconnect_info *vpninfo, struct oc_text_buf *h
 	return 0;
 }
 
-void cleanup_gssapi_auth(struct openconnect_info *vpninfo)
+void cleanup_gssapi_auth(struct openconnect_info *vpninfo,
+			 struct http_auth_state *auth_state)
+
 {
-	if (vpninfo->auth[AUTH_TYPE_GSSAPI].state >= AUTH_IN_PROGRESS) {
-		free(vpninfo->sspi_target_name);
-		vpninfo->sspi_target_name = NULL;
-		FreeCredentialsHandle(&vpninfo->sspi_cred);
-		DeleteSecurityContext(&vpninfo->sspi_ctx);
+	if (auth_state->state >= AUTH_IN_PROGRESS) {
+		free(auth_state->sspi_target_name);
+		auth_state->sspi_target_name = NULL;
+		FreeCredentialsHandle(&auth_state->sspi_cred);
+		DeleteSecurityContext(&auth_state->sspi_ctx);
 	}
 }
 
@@ -147,11 +150,12 @@ int socks_gssapi_auth(struct openconnect_info *vpninfo)
 	int first = 1;
 	int i;
 	int ret = -EIO;
+	struct http_auth_state *auth_state = &vpninfo->proxy_auth[AUTH_TYPE_GSSAPI];
 
-	if (sspi_setup(vpninfo, "rcmd"))
+	if (sspi_setup(vpninfo, auth_state, "rcmd", 1))
 		return -EIO;
 
-	vpninfo->auth[AUTH_TYPE_GSSAPI].state = AUTH_IN_PROGRESS;
+	vpninfo->proxy_auth[AUTH_TYPE_GSSAPI].state = AUTH_IN_PROGRESS;
 
 	pktbuf = malloc(65538);
 	if (!pktbuf)
@@ -172,12 +176,13 @@ int socks_gssapi_auth(struct openconnect_info *vpninfo)
 	out_token.pvBuffer = NULL;
 
 	while (1) {
-		status = InitializeSecurityContextW(&vpninfo->sspi_cred, first ? NULL : &vpninfo->sspi_ctx,
-						    vpninfo->sspi_target_name,
+		status = InitializeSecurityContextW(&auth_state->sspi_cred,
+						    first ? NULL : &auth_state->sspi_ctx,
+						    auth_state->sspi_target_name,
 						    ISC_REQ_ALLOCATE_MEMORY | ISC_REQ_CONFIDENTIALITY | ISC_REQ_REPLAY_DETECT | ISC_REQ_CONNECTION,
 						    0, SECURITY_NETWORK_DREP,
 						    first ? NULL : &input_desc,
-						    0, &vpninfo->sspi_ctx,
+						    0, &auth_state->sspi_ctx,
 						    &output_desc, &ret_flags, NULL);
 		if (status == SEC_E_OK) {
 			/* If we still have a token to send, send it. */
@@ -202,8 +207,7 @@ int socks_gssapi_auth(struct openconnect_info *vpninfo)
 
 		pktbuf[0] = 1; /* ver */
 		pktbuf[1] = 1; /* mtyp */
-		pktbuf[2] = (out_token.cbBuffer >> 8) & 0xff;
-		pktbuf[3] = out_token.cbBuffer & 0xff;
+		store_be16(pktbuf + 2, out_token.cbBuffer);
 		memcpy(pktbuf + 4, out_token.pvBuffer, out_token.cbBuffer);
 
 		FreeContextBuffer(out_token.pvBuffer);
@@ -236,7 +240,7 @@ int socks_gssapi_auth(struct openconnect_info *vpninfo)
 				     pktbuf[1]);
 			break;
 		}
-		in_token.cbBuffer = (pktbuf[2] << 8) | pktbuf[3];
+		in_token.cbBuffer = load_be16(pktbuf + 2);
 		in_token.pvBuffer = pktbuf;
 		first = 0;
 
@@ -267,7 +271,7 @@ int socks_gssapi_auth(struct openconnect_info *vpninfo)
 
 		ret = -EIO;
 
-		status = QueryContextAttributes(&vpninfo->sspi_ctx, SECPKG_ATTR_SIZES, &sizes);
+		status = QueryContextAttributes(&auth_state->sspi_ctx, SECPKG_ATTR_SIZES, &sizes);
 		if (status != SEC_E_OK) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("QueryContextAttributes() failed: %lx\n"), status);
@@ -303,7 +307,7 @@ int socks_gssapi_auth(struct openconnect_info *vpninfo)
 			goto err;
 		}
 
-		status = EncryptMessage(&vpninfo->sspi_ctx, SECQOP_WRAP_NO_ENCRYPT, &enc_desc, 0);
+		status = EncryptMessage(&auth_state->sspi_ctx, SECQOP_WRAP_NO_ENCRYPT, &enc_desc, 0);
 		if (status != SEC_E_OK) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("EncryptMessage() failed: %lx\n"), status);
@@ -331,8 +335,7 @@ int socks_gssapi_auth(struct openconnect_info *vpninfo)
 
 		pktbuf[0] = 1;
 		pktbuf[1] = 2;
-		pktbuf[2] = (len >> 8) & 0xff;
-		pktbuf[3] = len & 0xff;
+		store_be16(pktbuf + 2, len);
 
 		if (enc_bufs[0].cbBuffer)
 			memcpy(pktbuf + 4, enc_bufs[0].pvBuffer, enc_bufs[0].cbBuffer);
@@ -362,7 +365,7 @@ int socks_gssapi_auth(struct openconnect_info *vpninfo)
 			goto err;
 		}
 
-		len = (pktbuf[2] << 8) | pktbuf[3];
+		len = load_be16(pktbuf + 2);
 
 		i = vpninfo->ssl_read(vpninfo, (void *)pktbuf, len);
 		if (i < 0) {
@@ -385,7 +388,7 @@ int socks_gssapi_auth(struct openconnect_info *vpninfo)
 		enc_bufs[1].cbBuffer = 0;
 		enc_bufs[1].pvBuffer = NULL;
 
-		status = DecryptMessage(&vpninfo->sspi_ctx, &enc_desc, 0, NULL);
+		status = DecryptMessage(&auth_state->sspi_ctx, &enc_desc, 0, NULL);
 		if (status != SEC_E_OK) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("DecryptMessage failed: %lx\n"), status);
@@ -419,8 +422,8 @@ int socks_gssapi_auth(struct openconnect_info *vpninfo)
 	}
 
  err:
-	cleanup_gssapi_auth(vpninfo);
-	vpninfo->auth[AUTH_TYPE_GSSAPI].state = AUTH_UNSEEN;
+	cleanup_gssapi_auth(vpninfo, &vpninfo->proxy_auth[AUTH_TYPE_GSSAPI]);
+	vpninfo->proxy_auth[AUTH_TYPE_GSSAPI].state = AUTH_UNSEEN;
 	free(pktbuf);
 
 	return ret;
